# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import dash
from dash import dcc, html

import odatix.gui.navigation as navigation
import odatix.gui.themes as themes

import odatix.lib.printc as printc
from odatix.lib.utils import internal_error
import odatix.lib.term_mode as term_mode
from odatix.lib.settings import OdatixSettings

script_name = os.path.basename(__file__)
error_logfile = "odatix-gui_error.log"

class OdatixApp:
    def __init__(self,  old_settings=None, safe_mode=False, config_file=OdatixSettings.DEFAULT_SETTINGS_FILE, theme=themes.default_theme):
        
        # Get settings
        self.odatix_settings = OdatixSettings(config_file, silent=True)
        if not self.odatix_settings.valid:
            pass

        if theme is None:
            self.start_theme = themes.default_theme
        elif theme not in themes.list:
            printc.warning('Theme "' + str(theme) + '" does not exist. Using default theme.')
            self.start_theme = themes.default_theme
        else:
            self.start_theme = theme

        self.old_settings = old_settings
        self.safe_mode = safe_mode

        self.app = dash.Dash(
            name=__name__, 
            use_pages=True,
            title="Odatix",
            update_title="Odatix - Updating...",
            suppress_callback_exceptions=True
        )

        self.app.server.register_error_handler(Exception, self.handle_flask_exception)

        self.setup_layout()
        self.setup_callbacks()

    def handle_flask_exception(self, e):
        """
        Handle flask exceptions
        """
        if self.old_settings is not None:
            term_mode.restore_mode(self.old_settings)
        internal_error(e, error_logfile, script_name)
        if not self.safe_mode:
            os._exit(-1)

    def setup_layout(self):
        """
        Setup the layout of the Dash application.
        """
        self.app.layout = html.Div(
            children=[
                navigation.top_bar(self),
                navigation.side_bar(self),
                dcc.Location(id="url-global"),  
                dcc.Store(id="previous-url", data=""),
                dcc.Store(id="odatix-settings", data=self.odatix_settings.to_dict() if self.odatix_settings.valid else {}),
                html.Div(
                    [dash.page_container],
                    id="content",
                    className="content",
                    style={
                        "marginLeft": navigation.side_bar_width,
                        "width": "calc(100%-" + navigation.side_bar_width + ")",
                        "height": "100%",
                    },
                ),
            ],
            id="theme",
            style={
                "width": "100%",
                "height": "100%",
                "display": "flex",
                "flexDirection": "column"
            },
        )

    def setup_callbacks(self):
        """
        Setup Dash callbacks for interactivity.
        """
        navigation.setup_callbacks(self)


    def run(self):
        self.app.run(
            # host='0.0.0.0',
            host='127.0.0.1',
            debug=True
        )


if __name__ == "__main__":
    gui = OdatixApp()
    gui.run()



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import webbrowser
from threading import Thread
import socket
import logging 
import argparse
from waitress import serve

if sys.platform == "win32":
    import msvcrt
else:
    import select


import odatix.lib.printc as printc
import odatix.lib.term_mode as term_mode
from odatix.lib.settings import OdatixSettings
from odatix.gui.app import OdatixApp

sys.stdout = term_mode.RawModeOutputWrapper(sys.stdout)

######################################
# Settings
######################################

script_name = os.path.basename(__file__)

start_port = 8052
max_find_port_attempts = 50

######################################
# Parse Arguments
######################################

def add_arguments(parser):
    parser.add_argument('-i', '--input', type=str, default='results', help='Directory of the result YAML files')
    parser.add_argument('-n', '--network', action='store_true', help='Run the server on the network')
    parser.add_argument('-N', '--normal_term_mode', action='store_true', help='Do not change terminal mode')
    parser.add_argument('--safe_mode', action='store_true', help='Do not exit on internal error')
    parser.add_argument('-B', '--nobrowser', action='store_true', help='Do not open browser')
    parser.add_argument('-T', '--theme', default=None, help='Use a specific theme')
    parser.add_argument("-c", "--config", default=OdatixSettings.DEFAULT_SETTINGS_FILE, help="global settings file for Odatix (default: " + OdatixSettings.DEFAULT_SETTINGS_FILE + ")")

def parse_arguments():
    parser = argparse.ArgumentParser(description='Odatix - Start Result Explorer')
    add_arguments(parser)
    # Try enabling autocompletion if argcomplete is installed
    try:
        import argcomplete
        argcomplete.autocomplete(parser)
    except ImportError:
        pass  # No error if argcomplete is missing
    return parser.parse_args()

######################################
# Misc functions
######################################

def open_browser():
    webbrowser.open("http://" + ip_address + ':' + str(port), new=0, autoraise=True)

def close_server(old_settings):
    if old_settings is not None:
        term_mode.restore_mode(old_settings)
    os._exit(0)

def find_free_port(host, start_port):
    """Find a free port by incrementing from the start_port."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    port = start_port
    attempts = 0
    while True:
        if attempts >= max_find_port_attempts:
            printc.error(f"Could not find any available port in range [{start_port}:{port}]", script_name)
            sys.exit(-1)
        try:
            sock.bind((host, port))
            sock.close()
            return port
        except OSError:
            port += 1
            attempts += 1

def get_local_ip():
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
                s.connect(('8.8.8.8', 80))
                ip = s.getsockname()[0]
        except Exception:
                ip = '127.0.0.1'
        finally:
                s.close()
        return ip

def start_odatix_app(network=False, normal_term_mode=False, safe_mode=False, do_not_open_browser=False, config_file=OdatixSettings.DEFAULT_SETTINGS_FILE, theme=None):

    global ip_address
    global port

    # Default ip address: local
    host_address = '127.0.0.1'
    ip_address = host_address

    logging.getLogger('waitress').setLevel(logging.ERROR)

    if network:
        host_address = '0.0.0.0'
        ip_address = get_local_ip()

    port = find_free_port(host_address, start_port)

    printc.say("Server running on " + printc.colors.BLUE + "http://" + ip_address + ":" + str(port) + '/' + printc.colors.ENDC, end="", script_name=script_name)
    if network:
        print(" (network-accessible)")
    else:
        print(" (localhost only)")

    if normal_term_mode:
        printc.say("press 'q', then enter to quit", script_name=script_name)
    else:
        printc.say("press 'q' to quit", script_name=script_name)

    # Open the web page
    if not do_not_open_browser:
        process = Thread(target=open_browser).start()

    if normal_term_mode:
        old_settings = None
    else:
        old_settings = term_mode.set_raw_mode()

    odatix_app = OdatixApp(
        old_settings=old_settings,
        safe_mode=safe_mode,
        config_file=config_file,
        theme=theme,
    )

    # Start the server
    serve_thread = Thread(target=serve, args=(odatix_app.app.server,), kwargs={'host': host_address, 'port': port})
    serve_thread.start()
    
    try:
        while True:
            # Check if a key is pressed
            if sys.platform == "win32":
                if msvcrt.kbhit():
                    key = msvcrt.getch().decode("utf-8").lower()
                    if key == 'q':
                        close_server(old_settings)
            else:
                if sys.stdin in select.select([sys.stdin], [], [], 0.5)[0]:
                    key = sys.stdin.read(1).lower()
                    if key == 'q':
                        close_server(old_settings)
    finally:
        if old_settings is not None:
            term_mode.restore_mode(old_settings)

######################################
# Main
######################################

def main(args=None):
    if args is None:
        args = parse_arguments()

    network = args.network
    input = args.input
    normal_term_mode = args.normal_term_mode
    safe_mode = args.safe_mode
    do_not_open_browser = args.nobrowser
    config_file = args.config

    start_odatix_app(network, normal_term_mode, safe_mode, do_not_open_browser, config_file, args.theme)

if __name__ == "__main__":
    main()



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import dash
from dash import dcc, html

import odatix.explorer.navigation as navigation

dash.register_page(
  __name__,
  path='/',
  title='Odatix',
  name='Home',
)

cards = [
  {
    "name": "Lines / Points",
    "link": "/lines",
    "image": "assets/preview/lines.svg",
    "description": "A line chart to draw values for a specific metric",
  },
  {
    "name": "Columns",
    "link": "/columns",
    "image": "assets/preview/columns.svg",
    "description": "A column chart to draw values for a specific metric",
  },
  {
    "name": "Radar",
    "link": "/radar",
    "image": "assets/preview/radar.svg",
    "description": "A radar chart to draw values for a specific metric",
  },
  {
    "name": "Scatter",
    "link": "/scatter",
    "image": "assets/preview/scatter.svg",
    "description": "A scatter chart to draw a metric against another",
  },
  {
    "name": "Scatter 3D",
    "link": "/scatter3d",
    "image": "assets/preview/scatter3d.svg",
    "description": "A 3D scatter chart to draw a metric against another",
  },
  {
    "name": "Overview",
    "link": "/overview",
    "image": "assets/preview/overview.png",
    "description": "A collection of lines/columns/radar charts for all metrics",
  },
]

def create_button(page):
  return dcc.Link(
    html.Div(
      [
        html.Img(
          src=page["image"],
          className="card-img",
          style={"height": "300px"}
        ),
        html.Div(
          page["name"],
          className="card-title",
        ),
        html.Div(
          page["description"],
          className="card-description",
        ),
      ],
      className="card",
    ),
    href=page["link"],
    style={"text-decoration": "none"},
  )

padding = 20

layout = html.Div(
  [
    html.Div(
      [create_button(card) for card in cards],
      style={
        "display": "flex",
        "flex-wrap": "wrap",
        "justify-content": "center",
        "gap": "20px",
        "padding-left": "50px",
        "padding-right": "50px",
      },
    ),
  ],
  id=f"{__name__}-content",
  style={
    "width": "100%", 
    "min-height": f"calc(100vh - 1.8*{navigation.top_bar_height}",
    "display": "flex",
    "justify-content": "center",
    "align-items": "center",
    "padding-top": f"{padding}px",
    "padding-bottom": f"{padding}px",
  },
  className="home-content",
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#
import re
import sys
import argparse
from dash_svg import Svg, Path, Circle, Rect, Ellipse, Line, Polyline, Polygon
from xml.etree import ElementTree
from typing import Dict

######################################
# SVG to Inline Dash SVG
######################################

def add_arguments(parser: argparse.ArgumentParser):
    parser.add_argument("-i", "--input", default=None, help="input svg file")
    parser.add_argument("-o", "--output", default=None, help="input inline dash svg")

def parse_arguments():
    parser = argparse.ArgumentParser(description="Clean up current directory")
    add_arguments(parser)
    return parser.parse_args()

def kebab_to_camel(s: str) -> str:
    """
    Convert a kebab-case string to camelCase.
    Args:
        s (str): The input kebab-case string.
    """
    return re.sub(r'-([a-z])', lambda m: m.group(1).upper(), s)

def parse_style(style_str: str) -> Dict[str, str]:
    """
    Convert a CSS style string to a dictionary.
    Args:
        style_str (str): The CSS style string (e.g., "fill: #000; stroke: #fff;").
    Returns:
        dict: A dictionary representation of the styles (e.g., {"fill": "#000", "stroke": "#fff"}).
    """
    style_dict = {}
    for item in style_str.split(';'):
        if ':' in item:
            k, v = item.split(':', 1)
            k = kebab_to_camel(k.strip())
            style_dict[k] = v.strip()
    return style_dict

def normalize_attrib_keys(attrib: Dict[str, str]) -> Dict[str, str]:
    """
    Normalize SVG attribute names for Dash:
    - kebab-case to camelCase (e.g. clip-rule -> clipRule)
    - class -> className
    - keep data-* and aria-* as-is
    - strip namespaces
    """
    norm: Dict[str, str] = {}
    for k, v in attrib.items():
        # drop namespace prefix if present
        if "}" in k:
            k = k.split("}", 1)[1]
        # keep data-* / aria-* unchanged (Dash supports them)
        if k.startswith("data-") or k.startswith("aria-"):
            norm[k] = v
            continue
        if k == "class":
            norm["className"] = v
            continue
        norm[kebab_to_camel(k)] = v
    return norm

def svg_to_dashsvg(svg_file: str) -> Svg:
    """
    Convert an SVG file to a Dash SVG component.
    Args:
        svg_file (str): Path to the source SVG file.
    """
    tree = ElementTree.parse(svg_file)
    root = tree.getroot()
    children = []
    for elem in root.iter():
        tag = elem.tag.split('}')[-1] if '}' in elem.tag else elem.tag
        attrib = normalize_attrib_keys(dict(elem.attrib))   
        if 'style' in attrib and isinstance(attrib['style'], str):
            attrib['style'] = parse_style(attrib['style'])
        if tag.endswith('path'):
            children.append(Path(**attrib))
        elif tag.endswith('polygon'):
            children.append(Polygon(**attrib))
        elif tag.endswith('circle'):
            children.append(Circle(**attrib))
        elif tag.endswith('rect'):
            children.append(Rect(**attrib))
        elif tag.endswith('ellipse'):
            children.append(Ellipse(**attrib))
        elif tag.endswith('line'):
            children.append(Line(**attrib))
        elif tag.endswith('polyline'):
            children.append(Polyline(**attrib))
    svg_attrs = {}
    for k, v in root.attrib.items():
        if k.startswith('{'):
            continue
        key = k.split('}')[-1] if '}' in k else k
        key = kebab_to_camel(key)
        if key == "style":
            svg_attrs[key] = parse_style(v)
            pass
        else:
            svg_attrs[key] = v
    return Svg(children, **svg_attrs)

######################################
# Main
######################################

if __name__ == "__main__":
    args = parse_arguments()
    if args.input is None:
        print("Please provide an input svg file with -i/--input")
        sys.exit(1)
    code = svg_to_dashsvg(args.input)
    if args.output is None:
        print(code)
    else:
        with open(args.output, "w") as f:
            f.write(code)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State
import plotly.io as pio

import odatix.explorer.legend as legend
import odatix.explorer.themes as themes

from odatix.explorer.css_helper import Style

top_bar_height = "50px"
side_bar_width = "400px"

banned_pages = ["PageNotFound", "Home"]
sidebar_urls = ["/lines", "/columns", "/scatter", "/scatter3d", "/radar", "/overview"]

def top_bar(explorer):
  return html.Div(
    [
      dcc.Link(
        id="navbar-title",
        className="link",
        href="/",
        children=[
          html.Span(
            html.Div(
              [
                html.Span("Odatix Explorer", className="link-title1 title"),
                html.Span("Home", className="link-title2 title"),
              ],
              className="link-container"
            ),
            className="mask"
          )
        ],
        style={"position": "fixed", "margin-left": "30px", "left": "75px", "z-index": "2", "transition": "margin-left 0.25s"},
      ),
      html.Div([
        html.Div(
          [
            html.Div(
              dcc.Link(f"{page['name']}", href=page["relative_path"], className="nav-link"),
            ) for page in dash.page_registry.values() if page["name"] not in banned_pages
          ],
          className="nav-links",
        ),
        dcc.Dropdown(
          id="theme-dropdown",
          options=[{"label": f"{theme}", "value": f"{theme}"} for theme in reversed(list(themes.templates))],
          value=explorer.start_theme,
          className="theme-dropdown",
          clearable=False,
        ),
      ],
      id="nav-right",
      style={"display": "flex", "position": "absolute", "right": "0", "alignItems": "center", "justifyContent": "right", "z-index": "1000"},)
    ],
    style={"height": f"{top_bar_height}"},
    className="navbar",
    id="navbar",
  )


def side_bar(explorer):
  legend_items = legend.create_legend_items(explorer, "")
  target_legend_items = legend.create_target_legend_items(explorer, "")
  domain_legend_items = legend.create_domain_legend_items(explorer, "")

  return html.Div(
    [
      html.Div(
        id="sidebar-top",
        className="sidebar-top",
        style={"top": "0px", "left": "0px", "width": side_bar_width, "height": f"{top_bar_height}"},
      ),
      html.Img(
        id="toggle-button",
        className="sidebar-button",
        src="/assets/icons/sidebar-panel-expand-icon.svg",
        n_clicks=0,
        style={
          "display": "none",
          "cursor": "pointer",
          "position": "absolute",
          "top": "10px",
          "left": "20px",
          "width": "30px",
          "z-index": "3",
        },
      ),
      html.Div(
        id="sidebar",
        children=[
          html.Img(
            id="close-button",
            className="sidebar-button",
            src="/assets/icons/sidebar-panel-collapse-icon.svg",
            n_clicks=0,
            style={"cursor": "pointer", "position": "absolute", "top": "10px", "left": "20px", "width": "30px"},
          ),
          html.Div(children=[
            html.Div(
              id="sidebar-content",
              className="sidebar-content-holder",
              children=[
                html.H2("Data"),
                html.Div(
                  className="title-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("YAML File")]),
                    dcc.Dropdown(
                      id="yaml-dropdown",
                      options=[{"label": yaml_file, "value": yaml_file} for yaml_file in explorer.valid_yaml_files],
                      value=explorer.valid_yaml_files[0] if explorer.valid_yaml_files else None,
                      clearable=False,
                    ),
                  ],
                ),
                html.Div(
                  className="title-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Results")]),
                    dcc.Dropdown(
                      id="results-dropdown",
                      value="All",
                      options= ["All", "Fmax", "Custom Freq"],
                      clearable=False,
                    ),
                  ],
                ),
                html.Div(
                  className="title-dropdown",
                  id="title-metric-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Metric")]),
                    dcc.Dropdown(id="metric-dropdown", value="Fmax", clearable=False),
                  ],
                ),
                html.Div(
                  className="title-dropdown",
                  id="title-metric-x-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Metric X")]),
                    dcc.Dropdown(id="metric-x-dropdown", clearable=False),
                  ],
                ),
                html.Div(
                  className="title-dropdown",
                  id="title-metric-y-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Metric Y")]),
                    dcc.Dropdown(id="metric-y-dropdown", clearable=False),
                  ],
                ),
                html.Div(
                  className="title-dropdown",
                  id="title-metric-z-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Metric Z")]),
                    dcc.Dropdown(id="metric-z-dropdown", clearable=False),
                  ],
                ),
                html.Div(
                  id="overview-options",
                  children=[
                    html.H2("Overview Options"),
                    html.Div(
                      className="title-dropdown",
                      children=[
                        html.Div(className="dropdown-label", children=[html.Label("Chart Type")]),
                        dcc.Dropdown(
                          id="chart-type-dropdown",
                          options=[
                            {"label": "Lines", "value": "lines"},
                            {"label": "Columns", "value": "columns"},
                            {"label": "Radar", "value": "radar"}
                          ],
                          value="radar",
                          clearable=False,
                        ),
                      ],
                      style={"margin-bottom": "5px"},
                    ),
                    html.Div(
                      className="title-dropdown",
                      children=[
                        html.Div(className="dropdown-label", children=[html.Label("Layout")]),
                        dcc.Dropdown(
                          id="overview-layout-dropdown",
                          options=[
                            {"label": "Default", "value": "default"},
                            {"label": "Large", "value": "large"},
                            {"label": "Default (Tall)", "value": "default_tall"},
                            {"label": "Large (Tall)", "value": "large_tall"},
                            {"label": "Page Wide", "value": "page_wide"}
                          ],
                          value="default",
                          clearable=False,
                        ),
                      ],
                      style={"margin-bottom": "5px"},
                    ),
                  ],
                ),
                html.H2("Targets"),
                html.Div(
                  [
                    html.Div(
                      [
                        html.Button("Show All", id="show-all-targets", n_clicks=0),
                        html.Button("Hide All", id="hide-all-targets", n_clicks=0),
                      ]
                    ),
                    html.Div(target_legend_items, id="target-legend", style={"margin-top": "15px", "margin-bottom": "15px"}),
                  ],
                  style={"display": "inline-block", "margin-left": "20px"},
                ),
                html.H2("Architectures"),
                html.Div(
                  [
                    html.Div(
                      [
                        html.Button("Show All", id="show-all-architectures", n_clicks=0),
                        html.Button("Hide All", id="hide-all-architectures", n_clicks=0),
                      ]
                    ),
                    html.Div(legend_items, id="custom-legend", style={"margin-top": "15px", "margin-bottom": "15px"}),
                  ],
                  style={"display": "inline-block", "margin-left": "20px"},
                ),
                html.H2("Parameter Domains"),
                html.Div(
                  id="parameter-domains-error",
                  children=[
                    html.P(
                      [
                        "No parameter domain found in the result file.", html.Br(),
                        "Are these results from an older version of Odatix?"
                      ],
                      style={"margin-left": "20px", "margin-right": "20px"},
                    ),
                  ],
                ),
                html.Div(
                  id="parameter-domains",
                  children=[
                    html.Div(
                      className="title-dropdown",
                      id="title-dissociate-dropdown",
                      children=[
                        html.Div(className="dropdown-label", children=[html.Label("Dissociate Domain")]),
                        dcc.Dropdown(
                          id="dissociate-domain-dropdown",
                          options=[{"label": param, "value": param} for param in ["None"] + list(explorer.all_param_domains.keys())],
                          value="None",
                          placeholder="Domain",
                          clearable=False,
                        ),
                      ],
                      style={"margin-bottom": "5px"},
                    ),
                    html.Div(
                      className="title-dropdown",
                      children=[
                        html.Div(className="dropdown-label", children=[html.Label("Domain")]),
                        dcc.Dropdown(
                          id="param-domain-dropdown",
                          options=[{"label": param, "value": param} for param in explorer.all_param_domains.keys()],
                          value="__main__",
                          placeholder="Domain",
                          clearable=False,
                        ),
                      ],
                      style={"margin-bottom": "5px"},
                    ),
                    html.Div(
                      [
                        html.Div(
                          [
                            html.Button("Show All", id="show-all-domains", n_clicks=0),
                            html.Button("Hide All", id="hide-all-domains", n_clicks=0),
                          ]
                        ),
                        html.Div(domain_legend_items, id="domain-legend", style={"margin-top": "15px", "margin-bottom": "15px"}),
                      ],
                      style={"display": "inline-block", "margin-left": "20px"},
                    ),
                  ],
                ),
                html.H2("Display Settings"),
                html.Div(
                  className="title-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Color Mode")]),
                    dcc.Dropdown(
                      id="color-mode-dropdown",
                      options=[
                        {"label": "Target", "value": "target"},
                        {"label": "Architecture", "value": "architecture"},
                        {"label": "Domain Value", "value": "domain_value"},
                        {"label": "Frequency", "value": "frequency"},
                      ],
                      value="architecture",
                      clearable=False,
                    ),
                  ],
                  style={"margin-bottom": "5px"},
                ),
                html.Div(
                  className="title-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Symbol Mode")]),
                    dcc.Dropdown(
                      id="symbol-mode-dropdown",
                      options=[
                        {"label": "None", "value": "none"},
                        {"label": "Target", "value": "target"},
                        {"label": "Architecture", "value": "architecture"},
                        {"label": "Domain Value", "value": "domain_value"},
                        {"label": "Frequency", "value": "frequency"},
                      ],
                      value="target",
                      clearable=False,
                    ),
                  ],
                  style={"margin-bottom": "5px"},
                ),
                html.Div(
                  className="title-dropdown",
                      id="title-legend-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Legend Mode")]),
                    dcc.Dropdown(
                      id="legend-dropdown",
                      options=[
                        {"label": "Hide", "value": "hide_legend"},
                        {"label": "Show", "value": "show_legend"},
                        {"label": "Separate", "value": "separate_legend"},
                      ],
                      value="separate_legend",
                      clearable=False,
                    ),
                  ],
                  style={"margin-bottom": "5px"},
                ),
                html.Div(
                  className="toggle-container",
                  children=[
                    dcc.Checklist(
                      id="toggle-legend",
                      options=[{"label": " Show Legend", "value": True}],
                      value=[],
                      className="toggle",
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-legendgroup",
                      options=[{"label": " Show Legend Groups", "value": True}],
                      value=[True],
                      className="toggle",
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-title",
                      options=[{"label": " Show Title", "value": True}],
                      value=[True],
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-lines",
                      options=[{"label": " Show Lines", "value": True}],
                      value=[True],
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-lines-scatter",
                      options=[{"label": " Show Lines", "value": True}],
                      value=[],
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-close-line",
                      options=[{"label": " Close Lines", "value": True}],
                      value=[True],
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-connect-gaps",
                      options=[{"label": " Connect Gaps", "value": True}],
                      value=[True],
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-labels",
                      options=[{"label": " Show Labels", "value": True}],
                      value=[True],
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-zero-axis",
                      options=[{"label": " Start axis at zero", "value": True}],
                      value=[True],
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-unique-architectures",
                      options=[{"label": " Unique architectures color/symbol", "value": True}],
                      value=[True],
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                    dcc.Checklist(
                      id="toggle-unique-targets",
                      options=[{"label": " Unique targets color/symbol", "value": True}],
                      value=[],
                      labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                    ),
                  ],
                ),
                html.H2("Export Settings"),
                html.Div("Download in raster/vector format", style={"margin-left": "20px", "margin-bottom": "10px"}),
                html.Div(
                  className="title-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Download Format")]),
                    dcc.Dropdown(
                      id="dl-format-dropdown",
                      options=[
                        {"label": "SVG", "value": "svg"},
                        {"label": "PNG", "value": "png"},
                        {"label": "JPEG", "value": "jpeg"},
                        {"label": "WEBP", "value": "webp"},
                      ],
                      value="svg",
                      clearable=False,
                    ),
                  ],
                  style={"margin-bottom": "5px"},
                ),
                html.Div(
                  className="title-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Background Color")]),
                    dcc.Dropdown(
                      id="background-dropdown",
                      options=[
                        {"label": "Transparent", "value": "rgba(255, 255, 255, 0)"},
                        {"label": "White", "value": "rgba(255, 255, 255, 255)"},
                      ],
                      value="rgba(255, 255, 255, 0)",
                      clearable=False,
                    ),
                  ],
                ),
              ],
            )
          ], className="sidebar-content", style={"width": side_bar_width}),
        ],
        className="sidebar",
        style={"left": "0", "width": side_bar_width},
      ),
    ]
  )


def setup_sidebar_callbacks(explorer):
  @explorer.app.callback(
    [
      Output("title-metric-dropdown", "style"),
      Output("title-metric-x-dropdown", "style"),
      Output("title-metric-y-dropdown", "style"),
      Output("title-metric-z-dropdown", "style"),
      Output("title-legend-dropdown", "style"),
      Output("toggle-legend", "style"),
      Output("toggle-close-line", "style"),
      Output("toggle-lines", "style"),
      Output("toggle-lines-scatter", "style"),
      Output("toggle-connect-gaps", "style"),
      Output("toggle-zero-axis", "style"),
      Output("overview-options", "style"),
    ],
    [
      Input("url", "pathname"),
      Input("chart-type-dropdown", "value"),
    ],
  )
  def update_visibility(pathname, chart_type):

    if pathname in ["/scatter", "/scatter3d"]:
      dropdown_metric = Style.hidden
      dropdown_metric_x = Style.visible
      dropdown_metric_y = Style.visible
    elif pathname in ["/lines", "/columns", "/radar"]:
      dropdown_metric = Style.visible
      dropdown_metric_x = Style.hidden
      dropdown_metric_y = Style.hidden
    else:
      dropdown_metric = Style.hidden
      dropdown_metric_x = Style.hidden
      dropdown_metric_y = Style.hidden

    if pathname in ["/overview"]:
      toggle_legend = Style.hidden
      legend_dropdown = Style.visible
      overview_options = Style.visible_div
      pathname = "/" + chart_type
    else:
      toggle_legend = Style.visible
      legend_dropdown = Style.hidden
      overview_options = Style.hidden

    if pathname in ["/scatter3d"]:
      dropdown_metric_z = Style.visible
      toggle_zero_axis = Style.visible
    else:
      dropdown_metric_z = Style.hidden
      toggle_zero_axis = Style.hidden

    if pathname in ["/radar", "/overview"]:
      toggle_close_line = Style.visible
    else:
      toggle_close_line = Style.hidden

    if pathname in ["/columns"]:
      toggle_connect_gaps = Style.hidden
    else:
      toggle_connect_gaps = Style.visible

    if pathname in ["/columns"]:
      toggle_lines = Style.hidden
      toggle_lines_scatter = Style.hidden
    elif pathname in ["/scatter", "/scatter3d"]:
      toggle_lines = Style.hidden
      toggle_lines_scatter = Style.visible
    else:
      toggle_lines = Style.visible
      toggle_lines_scatter = Style.hidden

    return (
      dropdown_metric, dropdown_metric_x, dropdown_metric_y, dropdown_metric_z,
      legend_dropdown, toggle_legend, toggle_close_line, toggle_lines,
      toggle_lines_scatter, toggle_connect_gaps, toggle_zero_axis, overview_options
    )

  @explorer.app.callback(
    Output("parameter-domains", "style"),
    Output("parameter-domains-error", "style"),
    [
      Input("param-domain-dropdown", "value"),
      Input("yaml-dropdown", "value"),
    ],
  )
  def update_domain_checklist_states(selected_domain, selected_yaml):
    if len(explorer.param_domains[selected_yaml]) == 0:
      return Style.hidden, Style.visible_div
    return Style.visible_div, Style.hidden

  @explorer.app.callback(
    [
      Output("sidebar", "style"),
      Output("content", "style"),
      Output("navbar", "style"), 
      Output("navbar-title", "style"),
      Output("sidebar-top", "style"),
      Output("toggle-button", "style"),
      Output("previous-url", "data"),
    ],
    [
      Input("toggle-button", "n_clicks"),
      Input("close-button", "n_clicks"),
      Input("url", "pathname"),
      Input("theme-dropdown", "value"),
    ],
    [
      State("previous-url", "data"),
      State("sidebar", "style"),
      State("content", "style"),
      State("navbar", "style"),
      State("navbar-title", "style"),
      State("sidebar-top", "style"),
      State("toggle-button", "style"),
    ],
  )
  def toggle_sidebar(
    toggle_n_clicks, close_n_clicks, url, theme, previous_url, sidebar_style, content_style, navbar_style, navbar_title_style, sidebar_top, toggle_style
  ):
    ctx = dash.callback_context
    if not ctx.triggered:
      return sidebar_style, content_style, navbar_style, navbar_title_style, sidebar_top, toggle_style
    
    hide_button = False

    triggered_property = ctx.triggered[0]["prop_id"].split(".")[0]
    if triggered_property == "url":
      if url not in sidebar_urls:
        triggered_property = "close-button"
        hide_button = True
      else:
        if previous_url not in sidebar_urls:
          triggered_property = "toggle-button"

    if triggered_property == "toggle-button":
      if sidebar_style["left"] == "-" + side_bar_width:
        sidebar_style["left"] = "0"
        sidebar_top["left"] = "0"
        content_style["marginLeft"] = side_bar_width
        navbar_title_style["marginLeft"] = "30px"
        navbar_title_style["position"] = "fixed"
        toggle_style["display"] = "none"
      else:
        sidebar_style["left"] = "-" + side_bar_width
        sidebar_top["left"] = "-" + side_bar_width
        content_style["marginLeft"] = "0"
        navbar_title_style["marginLeft"] = "0"
        navbar_title_style["position"] = "relative"
        toggle_style["display"] = "block"
    elif triggered_property == "close-button":
      sidebar_style["left"] = "-" + side_bar_width
      sidebar_top["left"] = "-" + side_bar_width
      content_style["marginLeft"] = "0"
      navbar_title_style["marginLeft"] = "0"
      navbar_title_style["position"] = "relative"
      toggle_style["display"] = "block"

    if hide_button:
      toggle_style["display"] = "none"

    bar_background = themes.get_nav_bgcolor(theme)
    sidebar_style["backgroundColor"] = bar_background
    sidebar_top["backgroundColor"] = bar_background
    navbar_style["backgroundColor"] = bar_background
    navbar_title_style["backgroundColor"] = bar_background

    return sidebar_style, content_style, navbar_style, navbar_title_style, sidebar_top, toggle_style, url

  @explorer.app.callback(
    Output("main-container", "className"),
    Input("theme-dropdown", "value"),
  )
  def update_theme(
    theme
  ):
    return f"main-container {theme if theme != 'plotly' else ''}"



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

"""
This script allows replacing content between specified delimiters in a text file.

Functions:
    - parse_arguments: Parses command-line arguments.
    - read_file: Reads the contents of a file.
    - write_file: Writes content to a file.
    - replace_content: Replaces text between specified delimiters.
    - replace_params: Reads input, replaces content, and writes output.
    - main: Executes the script logic based on command-line arguments.
"""

import os
import re
import sys
import argparse
from typing import List

import odatix.lib.printc as printc
from odatix.lib.param_domain import ParamDomain

script_name = os.path.basename(__file__)

def add_arguments(parser):
    """Add command-line arguments."""
    parser.add_argument("-s", "--startdel", dest="start_delimiter", required=True, help="Start delimiter")
    parser.add_argument("-S", "--stopdel", dest="stop_delimiter", required=True, help="Stop delimiter")
    parser.add_argument("-i", "--input", dest="base_text_file", required=True, help="Input base text file")
    parser.add_argument("-o", "--output", dest="output_file", required=True, help="Output text file")
    parser.add_argument("-r", "--replace", dest="replacement_text_file", required=True, help="Replacement text file")
    parser.add_argument("-a", "--all", dest="replace_all_occurrences", action="store_true", help="Replace all occurrences")

def parse_arguments():
    """Parses command-line arguments."""
    parser = argparse.ArgumentParser(description="Replace content between delimiters in a text file.")
    add_arguments(parser)
    return parser.parse_args()

def read_file(file_path):
    """Reads the contents of a text file and returns it as a string."""
    try:
        with open(file_path, 'r') as file:
            return file.read()
    except Exception:
        printc.error("Could not open input file \"" + file_path + "\"", script_name)
        sys.exit(1)

def write_file(file_path, content):
    """Writes content to a specified text file."""
    try:
        with open(file_path, 'w') as file:
            file.write(content)
    except Exception as e:
        printc.error("Could not write output file \"" + file_path + "\": " + str(e), script_name)
        sys.exit(1)

def get_first_appearance(text, substring, start_line=1, start_char=1):
    """Returns the line and character number of the first appearance of a substring in the text."""
    lines = text.splitlines()
    for line_number, line in enumerate(lines[start_line-1:], start=start_line-1):
        start = start_char - 1 if line_number == start_line-1 else 0
        char_index = line.find(substring, start)
        if char_index != -1:
            return line_number + 1, char_index
    return -1, -1  # Not found

def replace_content(base_text, replacement_text, start_delim, stop_delim, replace_all_occurrences):
    """Replaces text between specified delimiters in the base text with the replacement text."""
    pattern = re.escape(start_delim) + '.*?' + re.escape(stop_delim)

    if start_delim == "" or stop_delim == "":
        return base_text, False
    
    match_found = re.search(pattern, base_text, flags=re.DOTALL) is not None
    
    if replace_all_occurrences:
        new_text = re.sub(pattern, start_delim + replacement_text + stop_delim, base_text, flags=re.DOTALL)
    else:
        new_text = re.sub(pattern, start_delim + replacement_text + stop_delim, base_text, count=1, flags=re.DOTALL)
    return new_text, match_found

def replace_params(base_text_file, replacement_text_file, output_file, start_delimiter, stop_delimiter, replace_all_occurrences=False, silent=False):
    """
    Reads the base file, replaces the text between delimiters by the text from the replacement file, and writes the updated content to the output file.

    Args:
        base_text_file (str): Path to the input text file where replacements will be made.
        replacement_text_file (str): Path to the file containing the replacement text.
        output_file (str): Path to the output text file where the modified content will be saved.
        start_delimiter (str): The starting delimiter marking the beginning of the replaceable section.
        stop_delimiter (str): The ending delimiter marking the end of the replaceable section.
        replace_all_occurrences (bool, optional): If True, replaces all occurrences of the pattern in the text. Defaults to False (only replaces the first occurrence).
        silent (bool, optional): If True, suppresses output messages. Defaults to False.
    """    
    # Read the contents of base and replacement files
    base_text = read_file(base_text_file)
    replacement_text = read_file(replacement_text_file)

    # Replace content between delimiters
    new_text, match_found = replace_content(base_text, replacement_text, start_delimiter, stop_delimiter, replace_all_occurrences)

    # Write the new content to the output file
    write_file(output_file, new_text)

    if not match_found:
        printc.warning("Could not find pattern ", script_name=script_name, end="")
        printc.red(start_delimiter, end="")
        printc.grey("[…]", end="")
        printc.red(stop_delimiter, end="")
        printc.yellow("\" in \"" + base_text_file + "\"")

    if not silent and match_found:
        if new_text != base_text:
            printc.say("Content replaced successfully, output saved to \"" + output_file + "\"", script_name)
        else:
            printc.say("Nothing to be done, input copied to \"" + output_file + "\"", script_name)

    return match_found

def replace_param_domain(param_domain: ParamDomain, design_path: str, silent: bool=False, debug: bool=False):
    """
    Replace parameters in a specific domain based on the provided ParamDomain object.

    Args:
        param_domain (ParamDomain): The ParamDomain object containing domain-specific information.
        design_path (str): The directory where the base file is located.
        silent (bool, optional): If True, suppresses output messages. Defaults to False.
        debug (bool, optional): If True, enables debug output. Defaults to False.
    """
    success = False
    if param_domain.use_parameters:
        param_target_file = os.path.join(design_path, param_domain.param_target_file)

        if debug: 
            printc.subheader("Replace parameters for \"" + param_domain.domain + "/" + param_domain.domain_value+ "\"")
        success = replace_params(
            base_text_file=param_target_file,
            replacement_text_file=param_domain.param_file,
            output_file=param_target_file,
            start_delimiter=param_domain.start_delimiter,
            stop_delimiter=param_domain.stop_delimiter,
            replace_all_occurrences=False,
            silent=False if debug else True,
        )
    return success

def replace_param_domains(param_domains: List[ParamDomain], design_path: str, timestamp=False, silent=False, debug=False):
    """
    Replace parameters for multiple ParamDomain objects and return a dictionary of domain values.

    Args:
        param_domains (List[ParamDomain]): A list of ParamDomain objects.
        design_path (str): The directory where the base file is located.
        timestamp (bool, optional): If True, includes a timestamp in the returned dictionary. Defaults to False.
        silent (bool, optional): If True, suppresses output messages. Defaults to False.
        debug (bool, optional): If True, enables debug output. Defaults to False.
    """
    domain_dict = {}
    if timestamp is not None:
        domain_dict["__timestamp__"] = timestamp 
    for param_domain in param_domains:
        success = replace_param_domain(param_domain, design_path, silent, debug)
        if success:
            domain_dict[param_domain.domain] = param_domain.domain_value
        if debug: 
            print()
    return domain_dict

######################################
# Main
######################################

def main(args):
    """Main function to handle argument parsing and execution of text replacement."""
    replace_params(args.base_text_file, args.replacement_text_file, args.output_file, args.start_delimiter, args.stop_delimiter, args.replace_all_occurrences)

if __name__ == "__main__":
    args = parse_arguments()
    main(args)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import dash
from dash import html, dcc, Input, Output, State, ctx
from typing import Optional#, Literal

import odatix.components.workspace as workspace
from odatix.gui.icons import icon
from odatix.gui.utils import get_key_from_url
import odatix.gui.ui_components as ui
import odatix.gui.navigation as navigation
import odatix.lib.hard_settings as hard_settings
from odatix.lib.settings import OdatixSettings

page_path = "/workspace"

dash.register_page(
    __name__,
    path=page_path,
    title='Odatix - Workspace Settings',
    name='Workspace',
    order=3,
)

######################################
# UI Components
######################################

def workspace_form_field(
    label: str,
    id: str,
    value: str="",
    tooltip: str="",
    placeholder: str="",
    tooltip_options: str="secondary",
    # type: Optional[Literal["text", "number", "password", "email", "range", "search", "tel", "url", "hidden"]] = None,
    type = None,
):
    return html.Div(
        children=[
            html.Label(label),
            ui.tooltip_icon(tooltip, tooltip_options),
            dcc.Input(id=id, value=value, type=type, placeholder=placeholder, style={"width": "100%"}),
        ],
        style={"marginBottom": "12px"}
    )

def workspace_form(settings):
    defval = lambda k, v=None: settings.get(k, v)
    use_benchmark = defval("use_benchmark", "")
    expand_design_path_filters = True if defval("design_path_whitelist", []) or defval("design_path_blacklist", []) else False

    return html.Div(
        children=[
            html.Div(style={"display": "none"}),
            html.Div([
                html.H3("Main Paths"),
                workspace_form_field(
                    label="Architecture directory",
                    id="arch_path",
                    value=defval("arch_path", ""),
                    placeholder=OdatixSettings.DEFAULT_ARCH_PATH,
                    tooltip="The path to the architecture directory where all architectures definitions are stored.",
                ),
                workspace_form_field(
                    label="Simulation directory",
                    id="sim_path",
                    value=defval("sim_path", ""),
                    placeholder=OdatixSettings.DEFAULT_SIM_PATH,
                    tooltip="The path to the architecture directory where all simulations definitions are stored.",
                ),
                workspace_form_field(
                    label="Target Definition directory",
                    id="target_path",
                    value=defval("target_path", ""),
                    placeholder=OdatixSettings.DEFAULT_TARGET_PATH,
                    tooltip="The path where target definitions for all tools are stored.",
                ),
            ], className="tile config"),
            html.Div([
                html.H3("Settings Files"),
                workspace_form_field(
                    label="Clean settings",
                    id="clean_settings_file",
                    value=defval("clean_settings_file", ""),
                    placeholder=OdatixSettings.DEFAULT_CLEAN_SETTINGS_FILE,
                    tooltip="The path to the clean rules file used by 'odatix clean' command.",
                ),
                workspace_form_field(
                    label="Simulation settings",
                    id="simulation_settings_file",
                    value=defval("simulation_settings_file", ""),
                    placeholder=OdatixSettings.DEFAULT_SIMULATION_SETTINGS_FILE,
                    tooltip="The path to the settings file used by 'odatix sim' command.",
                ),
                workspace_form_field(
                    label="Fmax synthesis settings",
                    id="fmax_synthesis_settings_file",
                    value=defval("fmax_synthesis_settings_file", ""),
                    placeholder=OdatixSettings.DEFAULT_FMAX_SYNTHESIS_SETTINGS_FILE,
                    tooltip="The path to the settings file used by 'odatix fmax' command.",
                ),
                workspace_form_field(
                    label="Custom synthesis settings",
                    id="custom_freq_synthesis_settings_file",
                    value=defval("custom_freq_synthesis_settings_file", ""),
                    placeholder=OdatixSettings.DEFAULT_CUSTOM_FREQ_SYNTHESIS_SETTINGS_FILE,
                    tooltip="The path to the settings file used by 'odatix freq' command.",
                ),
            ], className="tile config"),
            html.Div([
                html.H3("Work Directory"),
                workspace_form_field(
                    label="Main work path",
                    id="work_path",
                    value=defval("work_path", ""),
                    placeholder=OdatixSettings.DEFAULT_WORK_PATH,
                    tooltip="The path where all temporary files will be stored during synthesis and simulation.",
                ),
                workspace_form_field(
                    label="Simulation work path",
                    id="simulation_work_path",
                    value=defval("simulation_work_path", ""),
                    placeholder=OdatixSettings.DEFAULT_SIMULATION_WORK_PATH,
                    tooltip="The path where temporary files will be stored during simulations, relative to the main work path.",
                ),
                workspace_form_field(
                    label="Fmax synthesis work path",
                    id="fmax_synthesis_work_path",
                    value=defval("fmax_synthesis_work_path", ""),
                    placeholder=OdatixSettings.DEFAULT_FMAX_SYNTHESIS_WORK_PATH,
                    tooltip="The path where temporary files will be stored during Fmax synthesis, relative to the main work path.",
                ),
                workspace_form_field(
                    label="Custom synthesis work path",
                    id="custom_freq_synthesis_work_path",
                    value=defval("custom_freq_synthesis_work_path", ""),
                    placeholder=OdatixSettings.DEFAULT_CUSTOM_FREQ_SYNTHESIS_WORK_PATH,
                    tooltip="The path where temporary files will be stored during custom frequency synthesis, relative to the main work path.",
                ),
            ], className="tile config"),
            html.Div([
                html.H3("Results"),
                workspace_form_field(
                    label="Result path",
                    id="result_path",
                    value=defval("result_path", ""),
                    placeholder=OdatixSettings.DEFAULT_RESULT_PATH,
                    tooltip="The path where all results files will be stored after synthesis and simulation.",
                ),
                html.Div(
                    children=[
                        html.Label("Use benchmark results"),
                        ui.tooltip_icon(
                            "Include benchmark results in synthesis results.",
                            "secondary"
                        ),
                        dcc.Dropdown(
                            id="use_benchmark",
                            placeholder= f"{"Yes" if OdatixSettings.DEFAULT_USE_BENCHMARK else "No"}",
                            options=[
                                {"label": "Yes", "value": True},
                                {"label": "No", "value": False},
                            ],
                            value=use_benchmark,
                        ),
                    ],
                    style={"marginBottom": "12px"}
                ),
                html.Div(
                    children=[
                        workspace_form_field(
                            label="Benchmark file",
                            id="benchmark_file",
                            value=defval("benchmark_file", ""),
                            placeholder=OdatixSettings.DEFAULT_BENCHMARK_FILE,
                            tooltip="The path to the benchmark file.",
                        ),
                    ],
                    id="benchmark-file-container",
                    className="animated-section" + (" hide" if not use_benchmark else ""),
                    style={"overflow": "visible"},
                ),
            ], className="tile config"),
            
        ], className="tiles-container config", style={"marginTop": "-10px", "marginBottom": "20px"},
    )


######################################
# Callbacks
######################################

@dash.callback(
    Output("workspace-form-container", "children"),
    Output("workspace-initial-settings", "data"),
    Input(f"url_{page_path}", "search"),
    State(f"url_{page_path}", "pathname"),
)
def init_form(search, page):
    if page != page_path:
        return dash.no_update, dash.no_update

    settings = OdatixSettings.get_settings_file_dict()
    if settings is None:
        settings = {}
    return workspace_form(settings), settings

@dash.callback(
    Output({"page": page_path, "action": "save-all"}, "className"),
    Output({"page": page_path, "action": "save-all"}, "data-tooltip"),
    Output("odatix-settings", "data", allow_duplicate=True),
    Input({"page": page_path, "action": "save-all"}, "n_clicks"),
    Input("arch_path", "value"),
    Input("sim_path", "value"),
    Input("target_path", "value"),
    Input("clean_settings_file", "value"),
    Input("simulation_settings_file", "value"),
    Input("fmax_synthesis_settings_file", "value"),
    Input("custom_freq_synthesis_settings_file", "value"),
    Input("work_path", "value"),
    Input("simulation_work_path", "value"),
    Input("fmax_synthesis_work_path", "value"),
    Input("custom_freq_synthesis_work_path", "value"),
    Input("result_path", "value"),
    Input("use_benchmark", "value"),
    Input("benchmark_file", "value"),
    State(f"url_{page_path}", "pathname"),
    prevent_initial_call=True,
)
def save_and_status(
    save_n_clicks, arch_path, sim_path, target_path,
    clean_settings_file, simulation_settings_file, fmax_synthesis_settings_file,
    custom_freq_synthesis_settings_file, work_path, simulation_work_path,
    fmax_synthesis_work_path, custom_freq_synthesis_work_path, result_path,
    use_benchmark, benchmark_file,
    page
):
    triggered_id = ctx.triggered_id
    if triggered_id == f"url_{page_path}" and page != page_path:
        return dash.no_update, dash.no_update, dash.no_update
    
    current_settings_subset = {
        "arch_path": arch_path,
        "sim_path": sim_path,
        "target_path": target_path,
        "clean_settings_file": clean_settings_file,
        "simulation_settings_file": simulation_settings_file,
        "fmax_synthesis_settings_file": fmax_synthesis_settings_file,
        "custom_freq_synthesis_settings_file": custom_freq_synthesis_settings_file,
        "work_path": work_path,
        "simulation_work_path": simulation_work_path,
        "fmax_synthesis_work_path": fmax_synthesis_work_path,
        "custom_freq_synthesis_work_path": custom_freq_synthesis_work_path,
        "result_path": result_path,
        "use_benchmark": "" if use_benchmark is None else use_benchmark,
        "benchmark_file": benchmark_file,
    }
    
    current_settings = current_settings_subset

    if triggered_id == {"page": page_path, "action": "save-all"}:
        # Save settings
        try:
            OdatixSettings.save_dict_to_file(current_settings)
            current_settings = OdatixSettings().to_dict()
            return "color-button disabled icon-button tooltip delay bottom small", "Nothing to save", current_settings
        except Exception as e:
            return "color-button error-status icon-button tooltip bottom small", "Failed to save...", dash.no_update,

    else:
        settings = OdatixSettings.get_settings_template_dict()
        if current_settings_subset != settings:
            return "color-button warning icon-button tooltip bottom small tooltip", "Unsaved changes!", dash.no_update

    return "color-button disabled icon-button tooltip delay bottom small", "Nothing to save", dash.no_update


@dash.callback(
    Output("arch_path", "value"),
    Output("sim_path", "value"),
    Output("target_path", "value"),
    Output("clean_settings_file", "value"),
    Output("simulation_settings_file", "value"),
    Output("fmax_synthesis_settings_file", "value"),
    Output("custom_freq_synthesis_settings_file", "value"),
    Output("work_path", "value"),
    Output("simulation_work_path", "value"),
    Output("fmax_synthesis_work_path", "value"),
    Output("custom_freq_synthesis_work_path", "value"),
    Output("result_path", "value"),
    Output("use_benchmark", "value"),
    Output("benchmark_file", "value"),
    Input({"page": page_path, "action": "reset-defaults"}, "n_clicks"),
    State(f"url_{page_path}", "search"),
    State(f"url_{page_path}", "pathname"),
    State("workspace-initial-settings", "data"),
    State("workspace-saved-settings", "data"),
    State("odatix-settings", "data"),
    prevent_initial_call=True,
)
def reset_to_defaults(
     reset_n_clicks,
    search, page, initial_settings, saved_settings,
    odatix_settings,
):
    triggered_id = ctx.triggered_id
    if triggered_id == f"url_{page_path}" and page != page_path:
        return [dash.no_update] * 14
    
    if triggered_id == {"page": page_path, "action": "reset-defaults"}:
        # Reset to defaults
        return [""] * 14
    return [dash.no_update] * 14

@dash.callback(
    Output("benchmark-file-container", "className"),
    Input("use_benchmark", "value"),
)
def toggle_generate_settings(use_benchmark):
    return "animated-section" if use_benchmark else "animated-section hide"

######################################
# Layout
######################################

title_buttons = html.Div(
    children=[
        ui.icon_button(
            id={"page": page_path, "action": "reset-defaults"},
            icon=icon("reset", className="icon"),
            tooltip="Restore default workspace settings",
            tooltip_options="bottom",
            color="caution",
        ),
        ui.save_button(
            id={"page": page_path, "action": "save-all"},
            tooltip="Save all changes",
            disabled=True,
        ),
    ],
    className="inline-flex-buttons",
)

layout = html.Div(
    children=[
        dcc.Location(id=f"url_{page_path}"),
        html.Div(
            ui.title_tile(text="Workspace Settings", buttons=title_buttons, tooltip="Leave blank to use default values"), 
            id={"page": page_path, "type": "workspace-title-div"}, 
            style={"marginTop": "20px", "marginLeft": "-13px", "marginBottom": "10px"}
        ),
        html.Div(id="workspace-form-container"),
        dcc.Store(id="save-state", data=""),
        dcc.Store(id="workspace-initial-settings", data=None),
        dcc.Store(id="workspace-saved-settings", data=None), 
    ],
    className="page-content",
    style={
        "display": "flex",  
        "flexDirection": "column",
        "min-height": f"calc(100vh - {navigation.top_bar_height})",
    },
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import re
import time
import queue
import select
import signal
import subprocess
import curses
import locale
import io

if sys.platform == "win32":
  import msvcrt
  import time
  import ctypes
  import threading
  STD_BUF = ""
else:
  import fcntl
  STD_BUF = ""

from odatix.components.motd import read_version

from odatix.lib.curses_helper import enable_selection, disable_selection
from odatix.lib.ansi_to_curses import AnsiToCursesConverter
from odatix.lib.job_output_formatter import JobOutputFormatter
from odatix.lib.utils import open_path_in_explorer
import odatix.lib.printc as printc

ENCODING = locale.getpreferredencoding()

# Handle BUTTON5_PRESSED missing in some curses versions
if not hasattr(curses, "BUTTON5_PRESSED"):
  curses.BUTTON5_PRESSED = 2097152

if sys.platform == "win32":
  ENCODING = "utf-8"
  old_stdout = sys.stdout
  old_stderr = sys.stderr
  sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding=ENCODING, errors="replace")
  sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding=ENCODING, errors="replace")

######################################
# Settings
######################################

script_name = os.path.basename(__file__)
error = None

NORMAL = 1
RED = 2
YELLOW = 3
GREEN = 4
BLUE = 5
CYAN = 6
BLACK = 7
WHITE = 8
GREY = 9

REVERSE = 10
REVERSE_RED = 12
REVERSE_YELLOW = 13
REVERSE_GREEN = 14
REVERSE_BLUE = 15
REVERSE_CYAN = 16
REVERSE_BLACK = 17
REVERSE_WHITE = 18
REVERSE_GREY = 19

class Theme:
  theme = {
    'Color_Boxes': {
      'bar'               : '─',
      'border_left'       : ' ┊',
      'border_right'      : '┊ ',
      'progress_empty'    : '▅',
      'progress_full'     : '▅',
      'ballot_check'      : '➜ ',
      'ballot_empty'      : '  ',
      'spacer'            : ' ∙ ',
      'ellipsis'          : '…',
      'selected_bold'     : True,
      'selected_reverse'  : False,
      'colored_bar'       : True,
      'dim_empty_bar'     : True,
      'reserved_space'    : 32,
    },
    'ASCII_Highlight': {
      'bar'               : '-',
      'border_left'       : ' [',
      'border_right'      : '] ',
      'progress_empty'    : ' ',
      'progress_full'     : '#',
      'ballot_check'      : ' ',
      'ballot_empty'      : ' ',
      'spacer'            : '  ',
      'ellipsis'          : '...',
      'selected_bold'     : True,
      'selected_reverse'  : True,
      'colored_bar'       : False,
      'dim_empty_bar'     : False,
      'reserved_space'    : 29,
    },
    'ASCII_Highlight_Color': {
      'bar'               : '-',
      'border_left'       : ' [',
      'border_right'      : '] ',
      'progress_empty'    : ' ',
      'progress_full'     : '#',
      'ballot_check'      : ' ',
      'ballot_empty'      : ' ',
      'spacer'            : '  ',
      'ellipsis'          : '...',
      'selected_bold'     : True,
      'selected_reverse'  : True,
      'colored_bar'       : True,
      'dim_empty_bar'     : False,
      'reserved_space'    : 29,
    },
    'Color_Lines': {
      'bar'               : '─',
      'border_left'       : ' ┊',
      'border_right'      : '┊ ',
      'progress_empty'    : '━',
      'progress_full'     : '━',
      'ballot_check'      : ' ➜ ',
      'ballot_empty'      : '   ',
      'spacer'            : '  ',
      'ellipsis'          : '…',
      'selected_bold'     : True,
      'selected_reverse'  : False,
      'colored_bar'       : True,
      'dim_empty_bar'     : True,
      'reserved_space'    : 31,
    },
    'Legacy': {
      'bar'               : '─',
      'border_left'       : ' ┊',
      'border_right'      : '┊ ',
      'progress_empty'    : '🮏',
      'progress_full'     : '▅',
      'ballot_check'      : ' ✔ ',
      'ballot_empty'      : ' ❏ ',
      'spacer'            : '  ',
      'ellipsis'          : '…',
      'selected_bold'     : True,
      'selected_reverse'  : False,
      'colored_bar'       : False,
      'dim_empty_bar'     : False,
      'reserved_space'    : 31,
    },
    'Rectangles': {
      'bar'               : '─',
      'border_left'       : ' ┊',
      'border_right'      : '┊ ',
      'progress_empty'    : ' ',
      'progress_full'     : '▮',
      'ballot_check'      : ' ➜ ',
      'ballot_empty'      : '   ',
      'spacer'            : '  ',
      'ellipsis'          : '…',
      'selected_bold'     : True,
      'selected_reverse'  : False,
      'colored_bar'       : False,
      'dim_empty_bar'     : False,
      'reserved_space'    : 31,
    },
    'Rectangles_Color': {
      'bar'               : '─',
      'border_left'       : ' ┊',
      'border_right'      : '┊ ',
      'progress_empty'    : '▮',
      'progress_full'     : '▮',
      'ballot_check'      : ' ➜ ',
      'ballot_empty'      : '   ',
      'spacer'            : '  ',
      'ellipsis'          : '…',
      'selected_bold'     : True,
      'selected_reverse'  : False,
      'colored_bar'       : True,
      'dim_empty_bar'     : True,
      'reserved_space'    : 31,
    },
    'Simple': {
      'bar'               : '─',
      'border_left'       : ' ┊',
      'border_right'      : '┊ ',
      'progress_empty'    : '▭',
      'progress_full'     : '■',
      'ballot_check'      : ' ✔ ',
      'ballot_empty'      : ' ❏ ',
      'spacer'            : '  ',
      'ellipsis'          : '…',
      'selected_bold'     : True,
      'selected_reverse'  : False,
      'colored_bar'       : False,
      'dim_empty_bar'     : False,
      'reserved_space'    : 31,
    },
  }
  
  themes = list(theme.keys())

  def __init__(self, theme):
    if theme in Theme.themes:
      self.theme = theme
    else:
      raise ValueError(f"Unknown theme '{theme}'")

  def get(self, key):
    if self.theme not in Theme.themes:
      return '?'
    if key not in Theme.theme[self.theme]:
      return '?'
    return Theme.theme[self.theme][key]

  def next_theme(self):
    current_index = Theme.themes.index(self.theme)
    next_index = (current_index + 1) % len(Theme.themes)
    self.theme = Theme.themes[next_index]
    
######################################
# OS specific functions
######################################

def read_pipe_windows(pipe, job):
  while True:
    try:
      data = pipe.readline()
      if not data:
        break
      job.log_history.append(data)
      if job.log_size_limit != -1 and len(job.log_history) > job.log_size_limit:
        job.log_history = job.log_history[-job.log_size_limit:]
      job.log_changed = True
    except OSError:
      break

######################################
# ParallelJob
######################################

class ParallelJob:
  status_file_pattern = re.compile(r"(.*)")
  progress_file_pattern = re.compile(r"(.*)")

  def __init__(
    self,
    process,
    command,
    directory,
    generate_rtl,
    generate_command,
    target,
    arch,
    display_name,
    status_file,
    progress_file,
    tmp_dir,
    log_size_limit,
    progress_mode="default",
    status="not started",
  ):
    self.process = process
    self.command = command
    self.directory = directory
    self.generate_rtl = generate_rtl
    self.generate_command = generate_command
    self.target = target
    self.arch = arch
    self.display_name = display_name
    self.status_file = status_file
    self.progress_file = progress_file
    self.tmp_dir = tmp_dir
    self.log_size_limit = log_size_limit
    self.progress_mode = progress_mode
    self.status = status

    self.log_history = []
    self.log_position = 0
    self.log_changed = False
    self.autoscroll = True

    self.start_time = None
    self.stop_time = None

  @staticmethod
  def set_patterns(progress_file_pattern, status_file_pattern=None):
    ParallelJob.status_file_pattern = status_file_pattern
    ParallelJob.progress_file_pattern = progress_file_pattern

  def get_progress(self):
    if self.progress_mode == "fmax":
      return self.get_progress_fmax()
    else:
      progress = 0
      if os.path.isfile(self.progress_file):
        with open(self.progress_file, "r") as f:
          content = f.read()
        for match in re.finditer(ParallelJob.progress_file_pattern, content):
          parts = ParallelJob.progress_file_pattern.search(match.group())
          if len(parts.groups()) >= 2:
            progress = int(parts.group(2))
      if progress > 100:
        progress = 100
      return progress

  def get_progress_fmax(self):
    # Get progress from status file
    fmax_progress = 0
    fmax_step = 1
    fmax_totalstep = 1
    if os.path.isfile(self.status_file):
      with open(self.status_file, "r") as f:
        content = f.read()
      for match in re.finditer(ParallelJob.status_file_pattern, content):
        parts = ParallelJob.status_file_pattern.search(match.group())
        if len(parts.groups()) >= 4:
          fmax_progress = int(parts.group(2))
          fmax_step = int(parts.group(3))
          fmax_totalstep = int(parts.group(4))

    # Get progress from synth status file
    synth_progress = 0
    if os.path.isfile(self.progress_file):
      with open(self.progress_file, "r") as f:
        content = f.read()
      for match in re.finditer(ParallelJob.progress_file_pattern, content):
        parts = ParallelJob.progress_file_pattern.search(match.group())
        if len(parts.groups()) >= 2:
          synth_progress = int(parts.group(2))

    # Compute total progress
    if fmax_totalstep != 0:
      progress = fmax_progress + synth_progress / fmax_totalstep
    else:
      progress = synth_progress

    if progress > 100:
      progress = 100
    return progress

  def pause(self):
    """Suspend the job execution."""
    if self.process and self.status == "running":
      os.killpg(os.getpgid(self.process.pid), signal.SIGSTOP) # Suspend process
      self.status = "paused"
      self.stop_time = time.time()
      self.log_history.append(printc.colors.BLUE + "Job paused by user" + printc.colors.ENDC)

  def resume(self):
    """Resume the job execution."""
    if self.process and self.status == "paused":
      os.killpg(os.getpgid(self.process.pid), signal.SIGCONT) # Resume execution
      self.status = "running"
      self.start_time += time.time() - self.stop_time # Adjust the start time to reflect the paused duration
      self.stop_time = None
      self.log_history.append(printc.colors.GREEN + "Job resumed by user" + printc.colors.ENDC)

######################################
# ParallelJobHandler
######################################

class ParallelJobHandler:
  def __init__(self, job_list, nb_jobs=4, process_group=True, auto_exit=False, format_yaml=None, log_size_limit=200):
    self.job_list = job_list
    self.nb_jobs = nb_jobs
    self.process_group = process_group
    self.auto_exit = auto_exit
    self.log_size_limit = log_size_limit

    self.version = read_version()

    self.running_job_list = []
    self.retired_job_list = []
    self.job_queue = queue.Queue()
    self.selected_job_index = 0
    self.previous_log_size = 0
    self.max_title_length = max(len(job.display_name) for job in job_list)

    self.selection_enabled = False

    self.converter = AnsiToCursesConverter()
    if format_yaml is not None:
      self.formatter = JobOutputFormatter(format_yaml)
    else:
      self.formatter = None

    # Initial calculation of max displayed jobs
    height, _ = curses.initscr().getmaxyx()
    self.job_count = len(self.job_list)
    displayed_jobs_cmd = height // 2 - 2 # Default to half the remaining screen height
    max_displayed_jobs = min(self.job_count, displayed_jobs_cmd)
    
    # Job list
    self.job_index_start = 0
    self.job_index_end = max_displayed_jobs

    try:
      print(" ")
      self.theme = Theme('Color_Boxes')
    except:
      self.theme = Theme('ASCII_Highlight')

  @staticmethod
  def set_nonblocking(fd):
    if sys.platform != "win32":
      try:
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
      except TypeError:
        pass

  def progress_bar(self, window, id, progress, time, title, title_size, width, status="", selected=False):
    reserved_space = self.theme.get('reserved_space')
    spacer = self.theme.get('spacer')
    ellipsis = self.theme.get('ellipsis')

    if width > 100:
      display_time = True
    else:
      display_time = False
      reserved_space = reserved_space - (len(time) + len(spacer))

    bar_width = width - title_size - reserved_space
    if bar_width < 4:
      bar_width = 4
      title_size = width - bar_width - reserved_space
      if title_size < 0:
        title_size = 0

      if len(title) > title_size:
        title = title[: title_size - len(ellipsis)] + ellipsis
      else:
        title = title.ljust(title_size)
    else:
      title = title.ljust(title_size)

    bar_length = int(bar_width * progress / 100.0)

    real_id = self.job_index_start  + id

    try:
      # Style
      if real_id == self.selected_job_index and self.theme.get('selected_reverse') and self.job_count > 1:
        window.attron(curses.color_pair(NORMAL) | curses.A_REVERSE)
        attr = curses.A_REVERSE | curses.A_BOLD
        offset = REVERSE
      else:
        attr = 0
        offset = 0

      # Dim if popup menu opened
      attr = attr | curses.A_DIM if self.showing_help else attr

      # Ballot
      ballot = self.theme.get('ballot_check') if selected else self.theme.get('ballot_empty')
      window.addstr(id, 0, f"{ballot}")
      pos = len(ballot)
      
      # Title
      if real_id == self.selected_job_index and self.theme.get('selected_bold') and self.job_count > 1:
        window.attron(curses.color_pair(NORMAL) | curses.A_BOLD)
        attr = attr | curses.A_BOLD
      window.addstr(id, pos, f"{title}")
      window.attroff(curses.A_BOLD)
      pos = pos + len(title)

      # Border (left)
      border_left = self.theme.get('border_left')
      window.addstr(id, pos, f"{border_left}")
      pos = pos + len(border_left)

      # Progress bar (full)
      if self.theme.get('colored_bar'):
        if status == "failed" or status == "killed" or status == "canceled":
          color = curses.color_pair(RED + offset)
        elif status == "running":
          color = curses.color_pair(WHITE + offset)
        elif status == "success":
          color = curses.color_pair(GREEN + offset)
        elif status == "queued" or "paused":
          color = curses.color_pair(BLUE + offset)
        elif status == "starting":
          color = curses.color_pair(CYAN + offset)
        else:
          color = curses.color_pair(NORMAL + offset)
      else:
        color = 0
      window.addstr(id, pos, self.theme.get('progress_full') * bar_length, attr | color)
      
      # Progress bar (empty)
      if self.theme.get('dim_empty_bar'):
        dim = curses.color_pair(GREY + offset)
      else:
        dim = 0
      window.addstr(id, pos + bar_length, self.theme.get('progress_empty') * (bar_width - bar_length), attr | dim)
      pos = pos + bar_width

      # Border
      border_right = self.theme.get('border_right')
      window.addstr(id, pos, border_right + " ", attr)
      pos = pos + len(border_right)
      
      # Percentage
      percentage = f"{progress:3.0f}%"
      window.addstr(id, pos, percentage + spacer, attr)
      pos = pos + len(percentage) + len(spacer)
      
      # Elapsed time
      if display_time:
        window.addstr(id, pos, time + spacer, attr | curses.color_pair(GREY + offset))
        pos = pos + len(time) + len(spacer)

      # Status
      if status == "failed" or status == "killed" or status == "canceled":
        window.addstr(id, pos, status, curses.color_pair(RED + offset) | attr)
      elif status == "running":
        window.addstr(id, pos, status, curses.color_pair(YELLOW + offset) | attr)
      elif status == "success":
        window.addstr(id, pos, status, curses.color_pair(GREEN + offset) | attr)
      elif status == "queued" or "paused":
        window.addstr(id, pos, status, curses.color_pair(BLUE + offset) | attr)
      elif status == "starting":
        window.addstr(id, pos, status, curses.color_pair(CYAN + offset) | attr)
      else:
        window.addstr(id, pos, status, curses.color_pair(NORMAL + offset) | attr)
      pos = pos + len(status)

      # Padding
      remainder = width-pos
      if remainder > 0:
        window.addstr(id, pos, " "*(remainder), attr)
    except curses.error as e:
      pass
    
    window.attroff(curses.A_REVERSE)
    window.attroff(curses.A_BOLD)

  def update_header(self, header_win, active_jobs_count, retired_jobs_count, total_jobs_count, width):
    try:
      header_win.hline(0, 0, " ", width, curses.color_pair(NORMAL) | curses.A_REVERSE)
      header_win.addstr(0, 1, "v" + str(self.version), curses.color_pair(NORMAL) | curses.A_REVERSE)
    except curses.error:
      pass
    if total_jobs_count == 1:
      if retired_jobs_count == total_jobs_count:
        text = "{}/{} job done!".format(retired_jobs_count, total_jobs_count)
      else:
        text = "{}/{} job done - {} running".format(retired_jobs_count, total_jobs_count, active_jobs_count)
    else:
      if retired_jobs_count == total_jobs_count:
        text = "{}/{} jobs done!".format(retired_jobs_count, total_jobs_count)
      else:
        text = "{}/{} jobs done - {} running".format(retired_jobs_count, total_jobs_count, active_jobs_count)
    try:
      header_win.addstr(0, width - len(text) - 1, text, curses.color_pair(NORMAL) | curses.A_REVERSE)
    except curses.error:
      pass

    try:
      header_win.addstr(
        0, (width - len(" Odatix ")) // 2, " Odatix ", curses.color_pair(NORMAL) | curses.A_REVERSE | curses.A_BOLD
      )
    except curses.error:
      pass

    header_win.refresh()

  def update_progress_window(self, progress_win, selected_job):
    height, width = progress_win.getmaxyx()
    if self.showing_help:
      progress_win.attron(curses.A_DIM)
    
    # Clear the window
    progress_win.erase()

    # Display the jobs within the visible range
    for id, job in enumerate(self.job_list[self.job_index_start:self.job_index_end]):
      selected = (self.selected_job_index == self.job_index_start + id)
      if job.start_time is not None:
        if job.stop_time is not None:
          stop_time = job.stop_time
        else:
          stop_time = time.time()
        elapsed_seconds = int(stop_time - job.start_time)
        days = elapsed_seconds // 86400
        hours = (elapsed_seconds % 86400) // 3600
        minutes = (elapsed_seconds % 3600) // 60
        seconds = elapsed_seconds % 60
        if days > 0:
          elapsed_time = f"{days}d+{hours:02}:{minutes:02}" # Format Days+HH:MM
        else:
          elapsed_time = f"{hours:02}:{minutes:02}:{seconds:02}" # Format HH:MM:SS
      else:
        elapsed_time = "00:00:00"
      try:
        self.progress_bar(
          id=id,
          window=progress_win,
          progress=job.progress,
          time=elapsed_time,
          title=job.display_name,
          title_size=self.max_title_length,
          width=width,
          status=job.status,
          selected=selected,
        )
      except curses.error:
        pass
    progress_win.refresh()
    progress_win.attroff(curses.A_DIM)

  def update_separator(self, separator_win, val, ref, width):
    dim = self.showing_help and curses.A_DIM
    separator_win.erase()
    if val == ref:
      separator_text = self.theme.get("bar") * (width - 1)
    else:
      message = f"{val - ref} more" 
      padding = 4
      separator_text = self.theme.get("bar") * padding + message + self.theme.get("bar") * (width - len(message) - padding - 1)
    try:
      separator_win.addstr(0, 0, separator_text, dim)
    except curses.error as e:
      pass
    separator_win.refresh()

  @staticmethod
  def update_help(bottom_bar):
    bottom_bar.erase()

    # Define the text with attributes
    help_text = [
      ("q", "Quit"),
      ("h", "Help Menu"),
      ("c", "Cursor Mode"),
    ]

    bottom_bar.attron(curses.color_pair(NORMAL) | curses.A_REVERSE)
    try:
      bottom_bar.addstr(" ")
    except curses.error:
      pass

    for i, (key, description) in enumerate(help_text):
      try:
        if i > 0:
          bottom_bar.addstr(" | ")
        bottom_bar.attron(curses.color_pair(NORMAL) | curses.A_REVERSE | curses.A_BOLD)
        bottom_bar.addstr(key)
        bottom_bar.attroff(curses.A_BOLD)  # Remove attributes
        bottom_bar.addstr(": ")
        bottom_bar.addstr(description)
      except curses.error:
        pass

    try:
      bottom_bar.addstr(" ")
    except curses.error:
      pass
    bottom_bar.attroff(curses.color_pair(NORMAL) | curses.A_REVERSE | curses.A_BOLD)
    bottom_bar.refresh()

  @staticmethod
  def show_exit_confirmation(bottom_bar):
    try:
      bottom_bar.erase()
      bottom_bar.addstr(" Kill all jobs and exit: Yes (", curses.color_pair(NORMAL) | curses.A_REVERSE)
      bottom_bar.addstr("y", curses.color_pair(NORMAL) | curses.A_REVERSE | curses.A_BOLD)
      bottom_bar.addstr(") / No (", curses.color_pair(NORMAL) | curses.A_REVERSE)
      bottom_bar.addstr("n", curses.color_pair(NORMAL) | curses.A_REVERSE | curses.A_BOLD)
      bottom_bar.addstr(")? ", curses.color_pair(NORMAL) | curses.A_REVERSE)
      bottom_bar.refresh()
    except curses.error:
      pass

    key = bottom_bar.getch()
    curses.flushinp()
    if key == ord("y") or key == ord("Y"):
      return True, True
    elif key == ord("n") or key == ord("N"):
      return True, False
    else:
      return False, False

  @staticmethod
  def update_exit(bottom_bar):
    try:
      bottom_bar.erase()
      bottom_bar.addstr(" Exiting... ", curses.color_pair(NORMAL) | curses.A_REVERSE)
      bottom_bar.refresh()
    except curses.error:
      pass

  def update_logs(self, logs_win, selected_job, logs_height, width):
    history = selected_job.log_history
    log_length = len(history)

    # Erase extra lines from previous selected job
    if log_length < self.previous_log_size:
      for i in range(log_length, self.previous_log_size):
        try:
          logs_win.move(i, 0)
          logs_win.clrtoeol()
        except curses.error:
          pass

    self.converter.reset_format()

    # Logs from selected job
    for i, line in enumerate(history[selected_job.log_position : selected_job.log_position + logs_height]):
      try:
        logs_win.move(i, 0)
        if self.formatter is not None:
          line = self.formatter.replace_in_line(line)
        self.converter.add_ansi_str(logs_win, line, width=width, dim=self.showing_help)
      except curses.error:
        pass

    logs_win.refresh()
    self.previous_log_size = log_length

  def update_help_popup(self, popup_win, width, height, popup_width, popup_height):
    """Display a non-blocking help popup in the center of the screen while updating the other windows."""

    help_text = [
      ("q"          , "Quit"),
      ("PageUp    n", "Select next job"),
      ("PageDown  p", "Select previous job"),
      ("Up        u", "Scroll log"),
      ("Down      d", "Scroll log"),
      ("Home/End"   , "Scroll to top/bottom"),
      ("+/-"        , "Change progress window zize"),
      ("Space"      , "Pause selected job"),
      ("k"          , "Kill selected job"),
      ("s"          , "Start/resume selected job"),
      ("o"          , "Open current job work path"),
      ("t"          , "Switch theme"),
      ("h         ?", "Show help"),
    ]
    
    if popup_height <= 7:
      popup_win.erase()
    popup_win.box()
    
    try:
      close_x = popup_width - 5
      popup_win.addstr(1, close_x, " X ", curses.A_BOLD | curses.color_pair(REVERSE_RED))
    except curses.error:
      pass

    if popup_height > 2:
      try:
        text = " Help Menu "
        popup_win.addstr(1, (popup_width - len(text)) // 2, text, curses.A_BOLD | curses.A_REVERSE)
      except curses.error:
        pass
      if popup_height > 5:
        try:
          popup_win.addstr(3, 3, "Key      Alt", curses.A_BOLD )
          popup_win.addstr(3, 17, "Description", curses.A_BOLD )
          for i, (key, desc) in enumerate(help_text, start=4):
            if i < popup_height - 2:
              popup_win.addstr(i, 3, f"{key}", curses.color_pair(CYAN))
              popup_win.addstr(i, 17, f"{desc}")
        except curses.error:
          pass
    try:
      text = "Press 'q' or 'h' to close"
      popup_win.addstr(max(popup_height - 2, 1), (popup_width - len(text)) // 2, text, curses.A_DIM)
    except curses.error:
      pass
    popup_win.refresh()

  def start_job(self, job):
    # Run generate command
    if not job.generate_rtl:
      self.run_job(job)
      self.running_job_list.append(job)
      return

    try:
      job.log_history.append(printc.colors.CYAN + "Run generate command for " + job.display_name + printc.colors.ENDC)
      job.log_history.append(printc.colors.BOLD + " > " + job.generate_command + printc.colors.ENDC)

      process = subprocess.Popen(
        job.generate_command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE if sys.platform != "win32" else subprocess.STDOUT,
        cwd=job.tmp_dir,
        shell=True,
        universal_newlines=True,
        bufsize=1,
        encoding="utf-8",
        errors='replace',
      )

      self.set_nonblocking(process.stdout)
      self.set_nonblocking(process.stderr)

      job.process = process
      job.status = "starting"
      job.start_time = time.time()
      self.running_job_list.append(job)
    except subprocess.CalledProcessError:
      job.status = "failed"
      self.retire_job(job, progress=0)
      job.log_history.append(printc.colors.RED + "error: rtl generation failed" + printc.colors.ENDC)
      job.log_history.append(printc.colors.CYAN + "note: look for earlier error to solve this issue" + printc.colors.ENDC)
      return

  def click_on_job(self, progress_win, y, x):
    # Check if the click occured in progress_win
    if progress_win.enclose(y, x):
      # Get job index based on Y position
      relative_y = y - progress_win.getbegyx()[0]
      job_index = self.job_index_start + relative_y
      # Check if index is valid
      if 0 <= job_index < len(self.job_list):
        return job_index
    return -1

  def run_job(self, job):
    job.log_history.append(printc.colors.CYAN + "Run job command" + printc.colors.ENDC)
    job.log_history.append(printc.colors.BOLD + " > " + job.command + printc.colors.ENDC)

    preexec_fn = None
    if sys.platform != "win32" and self.process_group:
      preexec_fn = os.setpgrp

    process = subprocess.Popen(
      STD_BUF + job.command,
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE,
      cwd=job.directory,
      shell=True,
      universal_newlines=True,
      bufsize=1,
      preexec_fn=preexec_fn, 
      encoding="utf-8",
      errors='replace',
    )

    if job.start_time is None:
      job.start_time = time.time()
    job.process = process
    job.status = "running"
    
    if sys.platform == "win32":
      threading.Thread(target=read_pipe_windows, args=(process.stdout, job), daemon=True).start()
      threading.Thread(target=read_pipe_windows, args=(process.stderr, job), daemon=True).start()
    else:
      self.set_nonblocking(process.stdout)
      self.set_nonblocking(process.stderr)

  def queue_job(self, job):
    job.status = "queued"
    self.job_queue.put(job)

  def retire_job(self, job, progress=100):
    job.stop_time = time.time()
    self.running_job_list.remove(job)
    job.progress = progress
    self.retired_job_list.append(job)

  def terminate_all_jobs(self):
    for job in self.running_job_list:
      if job.process:
        try:  # Try to terminate the process group
          if sys.platform == "win32":
            job.process.send_signal(signal.CTRL_BREAK_EVENT)
          else:
            os.killpg(os.getpgid(job.process.pid), signal.SIGTERM)
        except ProcessLookupError:
          pass  # Process already terminated

    # Wait for all processes to finish
    for job in self.running_job_list:
      if job.process:
        job.process.wait()

  def read_process_output(self):
    if sys.platform == "win32":
      return
    
    pipes = [job.process.stdout for job in self.running_job_list] + [
      job.process.stderr for job in self.running_job_list
    ]
    
    try:
      read_ready, _, _ = select.select(pipes, [], [], 0.1)
    except:
      read_ready = []
    
    for pipe in read_ready:
      for job in self.running_job_list:
        job.log_changed = False
        if pipe in (job.process.stdout, job.process.stderr):
          while True:
            line = pipe.readline()
            if line:
              job.log_history.append(line)
              if self.log_size_limit != -1 and len(job.log_history) > self.log_size_limit:
                job.log_history = job.log_history[-self.log_size_limit:]
              job.log_changed = True
            else:
              break

  def curses_main(self, stdscr):
    curses.curs_set(0)  # Hide cursor

    # Enable mouse actions (disable selection)
    disable_selection()

    curses.start_color()
    curses.use_default_colors()

    # Define color pairs
    curses.init_pair(NORMAL, -1, -1)
    curses.init_pair(RED, curses.COLOR_RED, -1)
    curses.init_pair(YELLOW, curses.COLOR_YELLOW, -1)
    curses.init_pair(GREEN, curses.COLOR_GREEN, -1)
    curses.init_pair(BLUE, curses.COLOR_BLUE, -1)
    curses.init_pair(CYAN, curses.COLOR_CYAN, -1)
    curses.init_pair(BLACK, curses.COLOR_BLACK, -1)
    curses.init_pair(WHITE, curses.COLOR_WHITE, -1)
    curses.init_pair(GREY, curses.COLOR_BLACK + AnsiToCursesConverter.LIGHT_OFFSET, -1)

    curses.init_pair(REVERSE, curses.COLOR_BLACK, curses.COLOR_WHITE + AnsiToCursesConverter.LIGHT_OFFSET)
    curses.init_pair(REVERSE_RED, -1, curses.COLOR_RED + AnsiToCursesConverter.LIGHT_OFFSET)
    curses.init_pair(REVERSE_YELLOW, -1, curses.COLOR_YELLOW + AnsiToCursesConverter.LIGHT_OFFSET)
    curses.init_pair(REVERSE_GREEN, -1, curses.COLOR_GREEN + AnsiToCursesConverter.LIGHT_OFFSET)
    curses.init_pair(REVERSE_BLUE, -1, curses.COLOR_BLUE + AnsiToCursesConverter.LIGHT_OFFSET)
    curses.init_pair(REVERSE_CYAN, -1, curses.COLOR_CYAN + AnsiToCursesConverter.LIGHT_OFFSET)
    curses.init_pair(REVERSE_BLACK, -1, curses.COLOR_BLACK + AnsiToCursesConverter.LIGHT_OFFSET)
    curses.init_pair(REVERSE_WHITE, -1, curses.COLOR_WHITE + AnsiToCursesConverter.LIGHT_OFFSET)
    curses.init_pair(REVERSE_GREY, -1, curses.COLOR_BLACK)

    height, width = stdscr.getmaxyx()
    old_width = width
    old_height = height

    # Adjust window positions
    progress_height = self.job_index_end - self.job_index_start
    header_height = 1
    separator_height = 1
    help_height = 1
    logs_height = height - progress_height - 2*separator_height - help_height - header_height
    popup_width = min(50, width - 4)
    popup_height = min(19, height - 4)
    popup_height = max(popup_height, 3)
    start_x = (width - popup_width) // 2
    start_y = (height - popup_height) // 2

    try:
      header_win = curses.newwin(header_height, width, 0, 0)
      separator_top_win = curses.newwin(separator_height, width, header_height, 0)
      progress_win = curses.newwin(progress_height, width, header_height + separator_height, 0)
      separator_middle_win = curses.newwin(separator_height, width, header_height + separator_height + progress_height, 0)
      logs_win = curses.newwin(logs_height, width, header_height + separator_height + progress_height + separator_height, 0)
      bottom_bar = curses.newwin(help_height, width, height - help_height, 0)
      popup_win = curses.newwin(popup_height, popup_width, start_y, start_x)
      popup_win.box()
    except curses.error:
      stdscr.clear()
      try:
        stdscr.addstr(0, 0, "Could not start: window is too small. Press any key to exit", curses.color_pair(RED))
        stdscr.refresh()
      except curses.error:
        pass
      stdscr.getch()
      sys.exit(-1)

    finished = False
    ask_exit = False

    selected_job = self.job_list[self.selected_job_index]

    for job in self.job_list:
      if len(self.running_job_list) < self.nb_jobs:
        self.start_job(job)
      else:
        self.queue_job(job)

    total_jobs_count = len(self.job_list)

    stdscr.nodelay(True)

    resize = False
    resize_hold = False

    self.showing_help = False

    while True:
      height, width = stdscr.getmaxyx()
      # If window size changes, adjust the layout
      if height != old_height or width != old_width or resize:
        popup_width = min(50, width - 4)
        popup_height = min(20, height - 4)
        popup_height = max(popup_height, 3)
        start_x = (width - popup_width) // 2
        start_y = (height - popup_height) // 2
        try:
          progress_height = self.job_index_end - self.job_index_start
          logs_height = height - progress_height - 2*separator_height - help_height - header_height
          
          separator_top_win = curses.newwin(separator_height, width, header_height, 0)
          progress_win = curses.newwin(progress_height, width, header_height + separator_height, 0)
          separator_middle_win = curses.newwin(separator_height, width, header_height + separator_height + progress_height, 0)
          logs_win = curses.newwin(logs_height, width, header_height + separator_height + progress_height + separator_height, 0)
          bottom_bar = curses.newwin(help_height, width, height - help_height, 0)
          popup_win = curses.newwin(popup_height, popup_width, start_y, start_x)
          self.update_logs(logs_win, selected_job, logs_height, width)
        except curses.error:
          try:
            stdscr.clear()
            stdscr.addstr(0, 0, "Window is too small!", curses.color_pair(RED))
            stdscr.refresh()
          except curses.error:
            pass

        old_width = width
        old_height = height
        resize = False

      # Check if all jobs have finished
      active_jobs_count = len(self.running_job_list)
      queued_jobs_count = self.job_queue.qsize()
      if active_jobs_count == 0 and queued_jobs_count == 0:
        finished = True

      retired_jobs_count = len(self.retired_job_list)

      # Add a header
      self.update_header(header_win, active_jobs_count, retired_jobs_count, total_jobs_count, width)

      # Add a separator
      self.update_separator(separator_top_win, self.job_index_start, 0, width)

      # Update job status and progress
      for job in self.job_list:
        if job in self.retired_job_list:
          job.progress = job.progress
        elif job not in self.running_job_list or job.process is None:
          job.progress = 0
        else: 
          job.progress = job.get_progress()
          if job.process.poll() is not None:
            if job.process.returncode == 0:
              if job.status == "starting":
                job.log_history.append("")
                self.run_job(job)
                continue
              else:
                job.status = "success"
            else:
              job.status = "failed"
              if job.progress is None:
                job.progress = 0

            self.retire_job(job, job.progress)
            if not self.job_queue.empty():
              self.start_job(self.job_queue.get())
              
            if job == selected_job:
              selected_job.log_changed = True
              self.update_logs(logs_win, selected_job, logs_height, width)

      # Update progress window
      self.update_progress_window(progress_win, selected_job)

      # Add a separator
      self.update_separator(separator_middle_win, self.job_count, self.job_index_end, width)

      # Help window
      if self.showing_help:
        self.update_help_popup(popup_win, width, height, popup_width, popup_height)

      # Collect all stdout and stderr pipes
      self.read_process_output()

      # Automatically scroll if at the bottom
      if selected_job.autoscroll and selected_job.log_changed:
        selected_job.log_position = max(0, len(selected_job.log_history) - logs_height)
        self.update_logs(logs_win, selected_job, logs_height, width)

      # Handle resize
      if width != old_width:
        self.update_logs(logs_win, selected_job, logs_height, width)
      old_width = width

      # Get inputs
      key = stdscr.getch()
      curses.flushinp()    

      def update_selected_job():
        job = self.job_list[self.selected_job_index]
        job.log_position = max(0, len(job.log_history) - logs_height)
        job.autoscroll = True
        return job

      def scroll_up_logs(selected_job):
        if selected_job.log_position > 0:
          selected_job.log_position = max(0, selected_job.log_position - 3)
          selected_job.autoscroll = False
          self.update_logs(logs_win, selected_job, logs_height, width)

      def scroll_down_logs(selected_job):
        if selected_job.log_position + logs_height < len(selected_job.log_history):
          selected_job.log_position = min(len(selected_job.log_history) - logs_height, selected_job.log_position + 3)
          selected_job.autoscroll = False
          self.update_logs(logs_win, selected_job, logs_height, width)

      def scroll_up_progress():
        self.job_index_start -= 1
        self.job_index_end -= 1

      def scroll_down_progress():
        self.job_index_start += 1
        self.job_index_end += 1

      if not self.showing_help:

        if key == curses.KEY_MOUSE:
          _, x, y, _, button = curses.getmouse()

          if resize_hold:
            # Ensure the y-coordinate does not exceed the bounds of the terminal height
            y = min(y, height - 2)

            relative_y = y - (header_height + separator_height)
            new_job_index_end = self.job_index_start + relative_y

            # Minimum size
            if new_job_index_end <= self.job_index_start:
              self.job_index_end = self.job_index_start + 1
              resize = True
            # Exceeding job count
            elif new_job_index_end >= self.job_count:
              remainder = new_job_index_end - self.job_count
              self.job_index_end = self.job_count
              self.job_index_start = max(0, self.job_index_start - remainder)
              resize = True
            # General case
            else:
              self.job_index_end = new_job_index_end
              resize = True
        
          # Left click
          if button & curses.BUTTON1_CLICKED or button & curses.BUTTON1_DOUBLE_CLICKED:
            # Check if user left clicked on a job
            job_index = self.click_on_job(progress_win, y, x)
            # If user clicked on a job, update the selected job
            if job_index >= 0:
              self.selected_job_index = job_index
              selected_job = update_selected_job()
              self.update_logs(logs_win, selected_job, logs_height, width)
          
              # Check if it was a double click
              if button & curses.BUTTON1_DOUBLE_CLICKED:
                try:
                  open_path_in_explorer(self.job_list[job_index].tmp_dir)
                except NotImplementedError:
                  pass

          # Hold separator
          elif button & curses.BUTTON1_PRESSED:
            if separator_middle_win.enclose(y, x):
              resize_hold = True

          # Release
          elif button & curses.BUTTON1_RELEASED:
            resize_hold = False

          # Scroll up
          elif button & curses.BUTTON4_PRESSED:
            if progress_win.enclose(y, x):
              if self.job_index_start > 0:
                scroll_up_progress()
            elif logs_win.enclose(y, x):
              scroll_up_logs(selected_job)

          # Scroll down
          elif button & curses.BUTTON5_PRESSED:
            if progress_win.enclose(y, x):
              if self.job_index_end <= len(self.job_list) - 1:
                scroll_down_progress()
            elif logs_win.enclose(y, x):
              scroll_down_logs(selected_job)

        # Page Up
        elif key == curses.KEY_PPAGE or key == ord("p") or key == ord("P"):  
          if self.selected_job_index > 0:
            self.selected_job_index -= 1
          if self.selected_job_index < self.job_index_start:
            scroll_up_progress()
          selected_job = update_selected_job()
          self.update_logs(logs_win, selected_job, logs_height, width)

        # Page Down
        elif key == curses.KEY_NPAGE or key == ord("n") or key == ord("N"):
          if self.selected_job_index < len(self.job_list) - 1:
            self.selected_job_index += 1
          if self.selected_job_index >= self.job_index_end:
            scroll_down_progress()
          selected_job = update_selected_job()
          self.update_logs(logs_win, selected_job, logs_height, width)

        # Scroll Up Logs
        elif key == curses.KEY_UP or key == ord("u") or key == ord("U"):
          scroll_up_logs(selected_job)

        # Scroll Down
        elif key == curses.KEY_DOWN or key == ord("d") or key == ord("D"):
          scroll_down_logs(selected_job)

        # Logs Home
        elif key == curses.KEY_HOME:
          selected_job.log_position = 0
          selected_job.autoscroll = False
          self.update_logs(logs_win, selected_job, logs_height, width)

        # Logs End
        elif key == curses.KEY_END:
          selected_job.log_position = max(0, len(selected_job.log_history) - logs_height)
          selected_job.autoscroll = True
          self.update_logs(logs_win, selected_job, logs_height, width)

        # Expand progress window
        elif key == ord("+") or key == ord("="): 
          if logs_height > 0:
            if self.job_index_end < self.job_count:
              self.job_index_end += 1
              resize = True
            elif self.job_index_start > 0:
              self.job_index_start -= 1
              resize = True

        # Shrink progress window
        elif key == ord("-") or key == ord("_"):
          if self.selected_job_index < self.job_index_end - 1:
            self.job_index_end -= 1
            resize = True
          elif self.job_index_start < self.job_index_end - 1:
            self.job_index_start += 1
            resize = True

        # Enable/disable mouse actions/selection
        elif key == ord("c") or key == ord("C"):
          if self.selection_enabled:
            disable_selection()
            self.selection_enabled = False
          else:
            enable_selection()
            self.selection_enabled = True

        # Kill the selected job
        elif key == ord('k') or key == ord('K'):
          if selected_job.status == "running":  # Kill the running job
            try:
              if sys.platform == "win32":
                job.process.send_signal(signal.CTRL_BREAK_EVENT)
              else:
                os.killpg(os.getpgid(selected_job.process.pid), signal.SIGTERM)
              selected_job.status = "killed"
              self.retire_job(selected_job, selected_job.progress)
              selected_job.log_history.append(printc.colors.RED + "Job killed by user" + printc.colors.ENDC)
            except ProcessLookupError:
              selected_job.log_history.append(printc.colors.RED + "Failed to kill the job" + printc.colors.ENDC)
          elif selected_job.status == "queued":  # Remove job from queue
            try:
              self.job_queue.queue.remove(selected_job)
              selected_job.status = "canceled"
              selected_job.log_history.append(printc.colors.RED + "Job canceled by user" + printc.colors.ENDC)
            except ValueError:
              pass

        # Pause the selected job
        elif key == ord(' '):
          if selected_job.status == "running":
            selected_job.pause()

        # Start the selected job
        elif key == ord('s') or key == ord('S'): 
          if selected_job.status == "queued":
            self.job_queue.queue.remove(selected_job)
          if selected_job.status == "queued" or selected_job.status == "canceled":  # Start a queued job immediately
            self.start_job(selected_job)
          if selected_job.status == "paused":
            selected_job.resume()

        # Open job work directory in system file explorer
        elif key == ord('o') or key == ord('O'): 
          try:
            open_path_in_explorer(selected_job.tmp_dir)
          except NotImplementedError:
            pass

        # Change theme
        elif key == ord("t") or key == ord("T"):
          self.theme.next_theme()

        # Open help menu
        elif key in [ord("h"), ord("H"), ord("?")]:
          self.showing_help = True
          self.update_logs(logs_win, selected_job, logs_height, width) # Update to dim log window

        # Exit
        else:
          if self.auto_exit and finished:
            return True
          else:
            if key == ord("q") or key == ord("Q"):
              if finished:
                return True
              else:
                ask_exit = True
      
      else: # help window
        if key == curses.KEY_MOUSE:
          _, mouse_x, mouse_y, _, button = curses.getmouse()
          close_x = start_x + popup_width - 5
          close_y = start_y + 1

          if button & curses.BUTTON1_CLICKED and mouse_y == close_y and close_x <= mouse_x <= close_x + 2:
            self.showing_help = False
            logs_win.erase()
            self.update_logs(logs_win, selected_job, logs_height, width)

        elif key in [ord('h'), ord('H'), ord('?'), ord('q'), ord('Q')]:
          self.showing_help = False
          logs_win.erase() # Erase remains of help menu
          self.update_logs(logs_win, selected_job, logs_height, width)

      # Handle exit
      if ask_exit:
        user_answered, user_confirmed = self.show_exit_confirmation(bottom_bar)
        if user_answered:
          if user_confirmed:
            self.update_exit(bottom_bar)
            self.terminate_all_jobs()
            return False
          else:
            ask_exit = False
      else:
        self.update_help(bottom_bar)

  def run(self):
    curses.wrapper(self.curses_main)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import dash
import pandas as pd
from dash import dcc, html
from dash.dependencies import Input, Output
import plotly.graph_objs as go

import odatix.explorer.legend as legend
import odatix.explorer.content_lib as content_lib
import odatix.explorer.figures as figures
import odatix.explorer.themes as themes

def create_line_graph(
  explorer,
  selected_yaml,
  selected_metric,
  selected_results,
  show_all,
  hide_all,
  show_all_targets,
  hide_all_targets,
  show_all_domains,
  hide_all_domains,
  toggle_legend,
  toggle_legendgroup,
  toggle_title,
  toggle_lines,
  toggle_connect_gaps,
  toggle_labels,
  color_mode,
  symbol_mode,
  dl_format,
  background,
  theme,
  toggle_unique_architectures,
  toggle_unique_targets,
  dissociate_domain,
  arch_checklist_values,
  target_checklist_values,
  domain_checklist_values,
):

  fig = go.Figure()

  if not selected_yaml or selected_yaml not in explorer.dfs:
    return html.Div(className="error", children=[html.Div("Please select a YAML file.")])

  selected_metric_display = selected_metric.replace("_", " ") if selected_metric is not None else ""

  unit = legend.unit_to_html(explorer.units[selected_yaml].get(selected_metric, ""))
  selected_metric_display_unit = selected_metric_display + " (" + unit + ")" if unit else selected_metric_display

  ctx = dash.callback_context
  triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]

  if triggered_id in ["show-all-architectures", "hide-all-architectures"]:
    visible_architectures = set(
      explorer.dfs[selected_yaml]["Architecture"].unique() if triggered_id == "show-all-architectures" else []
    )
  else:
    visible_architectures = set(
      architecture for i, architecture in enumerate(explorer.all_architectures) if arch_checklist_values[i] and architecture in explorer.dfs[selected_yaml]["Architecture"].unique()
    )

  if triggered_id in ["show-all-targets", "hide-all-targets"]:
    visible_targets = set(
      explorer.dfs[selected_yaml]["Target"].unique() if triggered_id == "show-all-targets" else []
    )
  else:
    visible_targets = set(
      target for i, target in enumerate(explorer.all_targets) if target_checklist_values[i] and target in explorer.dfs[selected_yaml]["Target"].unique()
    )

  i_current_value = 0
  domains = {}
  for domain in explorer.all_param_domains.keys():
    domains[domain] = {}
    for config in explorer.all_param_domains[domain]:
      visible = True if domain_checklist_values[i_current_value] else False
      domains[domain].update({f'{config}': visible})
      i_current_value += 1


  filtered_df = explorer.dfs[selected_yaml][
    (explorer.dfs[selected_yaml]["Architecture"].isin(visible_architectures))
  ]

  for domain, configs in domains.items():
    if domain in filtered_df.columns:
      filtered_df.loc[:, domain] = filtered_df[domain].fillna("None")
      excluded_values = [config for config, is_visible in configs.items() if not is_visible]

      # Apply exclusion only if there are forbidden values in the column
      if any(filtered_df[domain].isin(excluded_values)):
        filtered_df = filtered_df[~filtered_df[domain].isin(excluded_values)]

  unique_configurations = sorted(filtered_df["Configuration"].unique())


  i_target = -1
  for i_unique_target, target in enumerate(explorer.all_targets):
    if target in explorer.dfs[selected_yaml]["Target"].unique():
      i_target = i_target + 1
    if target not in visible_targets:
      continue

    targets = [target] * len(unique_configurations)
    i_architecture = -1
    for i_unique_architecture, architecture in enumerate(explorer.all_architectures):
      if architecture in explorer.dfs[selected_yaml]["Architecture"].unique():
        i_architecture = i_architecture + 1
      if architecture not in visible_architectures:
        continue
        
      df_architecture_target = filtered_df[
        (filtered_df["Architecture"] == architecture) & 
        (filtered_df["Target"] == target)
      ]
      if df_architecture_target.empty:
        continue
      
      mode = "lines+markers" if toggle_lines else "markers"

      if dissociate_domain != "None" and dissociate_domain in df_architecture_target.columns:
        dissociation_values = df_architecture_target[dissociate_domain].dropna().unique()
        cleaned_configurations = [legend.clean_configuration_name(cfg, dissociate_domain) for cfg in unique_configurations]
      else:
        dissociation_values = [None]
        cleaned_configurations = unique_configurations

      i_dissociate_domain = -1
      for dissociation_value in dissociation_values:
        df_filtered = df_architecture_target.copy()
        if dissociation_value is not None and dissociation_value != "None":
          df_filtered = df_filtered[df_filtered[dissociate_domain] == dissociation_value]
          architecture_diplay = architecture + f" [{dissociate_domain.replace('__main__', 'main')}:{dissociation_value}]" 
        else:
          architecture_diplay = architecture
        i_dissociate_domain += 1
        
        df_fmax = df_filtered[df_filtered["Type"] == "Fmax"]
        if selected_results in ["All", "Fmax"] and not df_fmax.empty:
          if selected_metric is None or selected_metric not in df_fmax.columns:
            return html.Div(className="error", children=[html.Div("Please select a valid metric.")])

          y_values = [
            df_fmax[df_fmax["Configuration"] == config][selected_metric].values[0]
            if config in df_fmax["Configuration"].values
            else None
            for config in unique_configurations
          ]

          if color_mode == "architecture":
            color_id = i_unique_architecture if toggle_unique_architectures else i_architecture
          elif color_mode == "target":
            color_id = i_unique_target if toggle_unique_targets else i_target
          elif color_mode == "domain_value":
            color_id = i_dissociate_domain if dissociation_value != "None" else -1
          else:
            color_id = 0

          if symbol_mode == "none":
            symbol_id = 0
          elif symbol_mode == "architecture":
            symbol_id = i_unique_architecture if toggle_unique_architectures else i_architecture
          elif symbol_mode == "target":
            symbol_id = i_unique_target if toggle_unique_targets else i_target
          elif symbol_mode == "domain_value":
            symbol_id = i_dissociate_domain
          else:
            symbol_id = 0

          figures.add_trace_to_lines_fig(
            fig, cleaned_configurations, y_values, mode, architecture_diplay, "fmax",
            targets, target, selected_metric_display, unit, color_id, symbol_id, 
            toggle_legendgroup, toggle_connect_gaps
          )
          
        df_range = df_filtered[df_filtered["Type"] == "Custom Freq"]
        if selected_results in ["All", "Custom Freq"] and not df_range.empty:
          for i_freq, frequency in enumerate(explorer.all_frequencies):
            df_frequency = df_range[df_range["Frequency"] == frequency]
            if df_frequency.empty:
              continue

            if selected_metric is None or selected_metric not in df_frequency.columns:
              return html.Div(className="error", children=[html.Div("Please select a valid metric.")])

            if color_mode == "architecture":
              color_id = i_unique_architecture if toggle_unique_architectures else i_architecture
            elif color_mode == "target":
              color_id = i_unique_target if toggle_unique_targets else i_target
            elif color_mode == "domain_value":
              color_id = i_dissociate_domain if dissociation_value != "None" else -1
            else:
              color_id = i_freq + 1
              
            if symbol_mode == "none":
              symbol_id = 0
            elif symbol_mode == "architecture":
              symbol_id = i_unique_architecture if toggle_unique_architectures else i_architecture
            elif symbol_mode == "target":
              symbol_id = i_unique_target if toggle_unique_targets else i_target
            elif symbol_mode == "domain_value":
              symbol_id = i_dissociate_domain
            else:
              symbol_id = i_freq + 1

            y_values = [
              df_frequency[df_frequency["Configuration"] == config][selected_metric].values[0]
              if config in df_frequency["Configuration"].values
              else None
              for config in unique_configurations
            ]

            figures.add_trace_to_lines_fig(
              fig, cleaned_configurations, y_values, mode, architecture_diplay, f"{frequency} MHz",
              targets, target, selected_metric_display, unit, color_id, symbol_id, 
              toggle_legendgroup, toggle_connect_gaps
            )
  return fig, selected_metric_display, selected_metric_display_unit

def setup_callbacks(explorer, all_checklist_inputs, all_architecture_inputs, all_target_inputs, all_domains_inputs):

  @explorer.app.callback(
    Output("graph-lines", "children"),
    [
      Input("yaml-dropdown", "value"),
      Input("metric-dropdown", "value"),
      Input("results-dropdown", "value"),
      Input("show-all-architectures", "n_clicks"),
      Input("hide-all-architectures", "n_clicks"),
      Input("show-all-targets", "n_clicks"),
      Input("hide-all-targets", "n_clicks"),
      Input("show-all-domains", "n_clicks"),
      Input("hide-all-domains", "n_clicks"),
      Input("toggle-legend", "value"),
      Input("toggle-legendgroup", "value"),
      Input("toggle-title", "value"),
      Input("toggle-lines", "value"),
      Input("toggle-connect-gaps", "value"),
      Input("toggle-labels", "value"),
      Input("color-mode-dropdown", "value"),
      Input("symbol-mode-dropdown", "value"),
      Input("dl-format-dropdown", "value"),
      Input("background-dropdown", "value"),
      Input("theme-dropdown", "value"),
      Input("toggle-unique-architectures", "value"),
      Input("toggle-unique-targets", "value"),
      Input("dissociate-domain-dropdown", "value"),
    ]
    + all_checklist_inputs  
  )
  def update_line_graph(
    selected_yaml,
    selected_metric,
    selected_results,
    show_all,
    hide_all,
    show_all_targets,
    hide_all_targets,
    show_all_domains,
    hide_all_domains,
    toggle_legend,
    toggle_legendgroup,
    toggle_title,
    toggle_lines,
    toggle_connect_gaps,
    toggle_labels,
    color_mode,
    symbol_mode,
    dl_format,
    background,
    theme,
    toggle_unique_architectures,
    toggle_unique_targets,
    dissociate_domain,
    *checklist_values,
  ):
    if explorer is None:
      return html.Div(className="error", children=[html.Div("Loading...")])
    
    try:

      arch_checklist_values = checklist_values[:len(all_architecture_inputs)]
      target_checklist_values = checklist_values[len(all_architecture_inputs):len(all_architecture_inputs) + len(all_target_inputs)]
      domain_checklist_values = checklist_values[len(all_architecture_inputs) + len(all_target_inputs):]

      fig, selected_metric_display, selected_metric_display_unit = create_line_graph(
        explorer,
        selected_yaml,
        selected_metric,
        selected_results,
        show_all,
        hide_all,
        show_all_targets,
        hide_all_targets,
        show_all_domains,
        hide_all_domains,
        toggle_legend,
        toggle_legendgroup,
        toggle_title,
        toggle_lines,
        toggle_connect_gaps,
        toggle_labels,
        color_mode,
        symbol_mode,
        dl_format,
        background,
        theme,
        toggle_unique_architectures,
        toggle_unique_targets,
        dissociate_domain,
        arch_checklist_values,
        target_checklist_values,
        domain_checklist_values,
      )
      
      fig.update_layout(
        paper_bgcolor=background,
        showlegend=True if toggle_legend else False,
        legend_groupclick="toggleitem",
        xaxis_title="Configuration",
        yaxis_title=selected_metric_display_unit,
        yaxis=dict(range=[0, None]),
        title=selected_metric_display if toggle_title else None,
        title_x=0.5,
        autosize=True,
        template=theme,
        modebar={
          "bgcolor": themes.get_page_bgcolor(theme, default=None),
          "color": themes.get_button_color(theme, default=None),
          "activecolor": themes.get_button_active_color(theme, default=None),
        }
      )

      fig.update_xaxes(
        showticklabels=True if toggle_labels else False
      )

      filename = "Odatix-{}-{}-{}".format(
        os.path.splitext(selected_yaml)[0], __name__, selected_metric
      )
      
      return html.Div(
        [
          dcc.Graph(
            figure=fig,
            style={"width": "100%", "height": "100%"},
            config={
              "displayModeBar": True,
              "displaylogo": False,
              "scrollZoom": True,
              "modeBarButtonsToRemove": ["lasso", "select"],
              "toImageButtonOptions": {
                "format": dl_format,
                "scale": 3,
                "filename": filename
              },
            },
          )
        ],
        style={"width": "100%", "height": "100%", "display": "inline-block", "vertical-align": "top"},
      )
    except Exception as e:
      return content_lib.generate_error_div(e)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import sys

if sys.platform != "win32":
  import tty
  import termios

def set_raw_mode():
  """
  Sets the terminal to raw mode, which allows reading input byte by byte without requiring Enter.
  This function saves the current terminal settings so they can be restored later.
  
  Returns:
    old_settings (list): The original terminal settings before switching to raw mode.
  """
  if sys.platform != "win32":
    fd = sys.stdin.fileno()
    global old_settings
    old_settings = termios.tcgetattr(fd)
    tty.setraw(fd)
    return old_settings

def restore_mode(old_settings):
  """
  Restores the terminal to a previously saved state.

  Args:
    old_settings (list): The terminal settings to restore.
  """
  if sys.platform != "win32":
    fd = sys.stdin.fileno()
    termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

class RawModeOutputWrapper:
  """
  A wrapper for sys.stdout that ensures compatibility with raw terminal mode.
  
  In raw mode, a newline character (`\n`) only moves the cursor to the next line
  but does not return it to the beginning of the line. This class intercepts all 
  output to sys.stdout and replaces every `\n` with the combination `\r\n`, 
  which moves the cursor both to the next line and back to the beginning.
  """
  def __init__(self, wrapped):
    self.wrapped = wrapped # The original sys.stdout or any file-like object to wrap/

  def write(self, text):
    """ 
    Replaces `\n` with `\r\n` and writes the modified text to the wrapped output
    """
    text = text.replace("\n", "\r\n")
    self.wrapped.write(text)

  def flush(self):
    """
    Flushes the wrapped output stream
    """
    self.wrapped.flush()



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import yaml
import math
import itertools
from typing import Optional

import odatix.lib.printc as printc
import odatix.lib.hard_settings as hard_settings
from odatix.lib.get_from_dict import get_from_dict, Key

script_name = os.path.basename(__file__)

dimension_types = ("bool", "range", "list", "multiples", "power_of_two")
modification_types = ("union", "disjunctive_union", "intersection", "difference")
combo_types = ("function", "conversion", "format")

######################################
# Generator
######################################

class ConfigGenerator:
  """
  Generates parameter configurations based on a YAML file defining variable types,
  ranges, and naming templates.
  """

  def __init__(self, path: str = "", data: Optional[dict] = None, silent: bool = False, debug: bool = False):
    """
    Initialize the configuration generator.
    
    Args:
        path (str): Path to the directory containing the "_settings.yml" file.
        data (dict): Optional pre-loaded YAML data to use instead of reading from file.
        silent (bool): If True, suppress warnings for missing keys.
        debug (bool): If True, print debug information.
    """
    self.path = path
    self.silent = silent
    self.debug = debug

    self.template = ""
    self.name_template = ""
    self.variables = {}
    self.valid = False
    self.enabled = False
    self.error_messages = []

    if data is not None:
      self.yaml_file = "<provided_data>"
      self.data = data
    else:
      self.yaml_file = os.path.join(self.path, hard_settings.param_settings_filename)
      self._load_yaml()

    self._validate_keys(self.data)

  def _load_yaml(self):
    """
    Load YAML settings and validate the presence of necessary keys.
    
    Returns:
        dict: Parsed YAML data.
    """
    self.data = {}
    if not os.path.isfile(self.yaml_file):
      printc.error(f"YAML file '{self.yaml_file}' does not exist.")
      sys.exit(-1)

    try:
      with open(self.yaml_file, "r") as f:
        data = yaml.safe_load(f)
        if data is None:
          printc.error(f"YAML file \"{self.yaml_file}\" is empty!")
          sys.exit(-1)
    except yaml.YAMLError as e:
      printc.error(f"Invalid YAML file \"{self.yaml_file}\": {e}")
      sys.exit(-1)
    self.data = data

  def _validate_keys(self, data):
    # Check main keys
    hide = not self.debug
    generate_enabled, generate_defined = get_from_dict("generate_configurations", data, self.yaml_file, default_value=False, silent=hide, script_name=script_name)
    generate_settings, generate_settings_defined = get_from_dict("generate_configurations_settings", data, self.yaml_file, silent=hide, script_name=script_name)
    
    if generate_settings_defined:
      self.template, template_defined = get_from_dict("template", generate_settings, self.yaml_file, parent="generate_configurations_settings", behavior=Key.MANTADORY, script_name=script_name)
      self.name_template, name_template_defined = get_from_dict("name", generate_settings, self.yaml_file, parent="generate_configurations_settings", type=str, behavior=Key.MANTADORY, script_name=script_name)
      self.variables, variables_defined = get_from_dict("variables", generate_settings, self.yaml_file, parent="generate_configurations_settings", type=dict, behavior=Key.MANTADORY, script_name=script_name)
      self.valid = generate_settings_defined and template_defined and name_template_defined and variables_defined and generate_defined
    else:
      self.valid = False

    # Concat all strings if template is a list
    if isinstance(self.template, list):
      self.template = "\n".join(map(str, self.template)) 

    if not generate_defined and generate_settings_defined and not self.silent:
      printc.warning('"generate_configurations_settings" is defined while "generate_configurations" is not. Disabling configuration generation.', script_name)
    if generate_defined and generate_enabled and not generate_settings_defined and not self.silent:
      printc.error('Configuration generation is enabled while "generate_configurations_settings" is not defined.', script_name)

    self.enabled = generate_enabled

  def evaluate_expression(self, expr, values_map):
    """
    Evaluate a mathematical expression using the current values of variables.

    Args:
        expr (str): The expression to evaluate.
        values_map (dict): A dictionary containing values for referenced variables.

    Returns:
        Any: The result of evaluating the expression.
    """
    safe_env = {var: values_map[var] for var in values_map}
    safe_env["math"] = math
    expr = expr.replace("^", "**")  # Replace '^' with '**' for Python exponentiation
    expr = expr.replace("$", "")
    expr = expr.replace("{", "")
    expr = expr.replace("}", "")
    try:
      return eval(expr, {"__builtins__": None}, safe_env)
    except Exception as e:
      printc.error(f"Failed to evaluate expression '{expr}': {e}", script_name)
      return None

  def generate(self):
    """
    Generate all possible parameter combinations based on the configuration.
    Includes support for union, disjunctive_union, intersection, and difference.
    
    Returns:
        dict: A dictionary where keys are generated names and values are formatted templates.
        dict: A dictionary where keys are variable names and values are lists of all possible values for those variables.
    """
    if not self.valid or not self.enabled:
      return {}

    sources_used = set()
    for variable, config in self.variables.items():
      value_type, value_type_defined = get_from_dict("type", config, self.yaml_file, parent=variable, behavior=Key.MANTADORY, script_name=script_name)
      if value_type in modification_types:
        settings, settings_defined = get_from_dict("settings", config, self.yaml_file, parent=variable, behavior=Key.MANTADORY, script_name=script_name)
        if not settings_defined:
          return {}
        sources, sources_defined = get_from_dict("sources", settings, self.yaml_file, parent=f"{variable}[settings]", behavior=Key.MANTADORY, type=list, script_name=script_name)
        if not sources_defined:
          return {}
        for source in sources:
          sources_used.add(source)
      elif value_type not in dimension_types and value_type not in combo_types:
        printc.error(f'Invalid type \"{value_type}\" for variable "{variable}", in ' + self.yaml_file + '".', script_name)
        return {}

    dimension_vars = {}
    for variable, config in self.variables.items():
      value_type, _ = get_from_dict("type", config, self.yaml_file, parent=variable, behavior=Key.MANTADORY, script_name=script_name)
      if value_type in dimension_types and variable not in sources_used:
        dimension_vars[variable] = self.generate_values_for_dim(variable, config)

    result_set = set()
    for variable, config in self.variables.items():
      value_type, _ = get_from_dict("type", config, self.yaml_file, parent=variable, behavior=Key.MANTADORY, script_name=script_name)
      if value_type in modification_types:
        settings, settings_defined = get_from_dict("settings", config, self.yaml_file, behavior=Key.MANTADORY, script_name=script_name)
        if settings_defined:
          sources, sources_defined = get_from_dict("sources", settings, self.yaml_file, behavior=Key.MANTADORY, default_value=[], type=list, script_name=script_name)
          if sources_defined:
            sources = [source.replace("$", "").replace("{", "").replace("}", "") for source in sources]
            sets = [set(self.generate_values_for_dim(source, self.variables.get(source, {}))) for source in sources if source in self.variables]
            
            if value_type == "union":
              if sets:
                result_set = set.union(*sets)
              else:
                result_set = set()
            elif value_type == "disjunctive_union":
              if sets:
                result_set = set.union(*sets) - set.intersection(*sets)
              else:
                result_set = set()
            elif value_type == "intersection":
              if sets:
                result_set = set.intersection(*sets)
              else:
                result_set = set()
            elif value_type == "difference":
              if len(sets) == 2:
                result_set = sets[0] - sets[1]
              else:
                result_set = set()
                printc.error(f'{variable} -> The\"difference\" operation only supports \"sources\" having exactly two elements, in ' + self.yaml_file + '".', script_name)       
            else:
              result_set = set()
              printc.warning(f'Invalid operation for variable "{variable}", in ' + self.yaml_file + '".', script_name)

            dimension_vars[variable] = sorted(result_set)

    if self.debug:
      printc.note(f"dimension_vars after set operations: {dimension_vars}", script_name)

    var_names = list(dimension_vars.keys())
    combos = list(itertools.product(*(dimension_vars[k] for k in var_names)))
    final_configs = {}
    all_vars_values= {}
    values = set()

    for combo in combos:
      value_map = dict(zip(var_names, combo))
      for variable, config in self.variables.items():
        value_type, _ = get_from_dict("type", config, self.yaml_file, parent=variable, behavior=Key.MANTADORY, script_name=script_name)
        if value_type == "function":
          settings, defined = get_from_dict("settings", config, self.yaml_file, silent=True, script_name=script_name)
          if defined:
            op, _ = get_from_dict("op", settings, self.yaml_file, silent=True, script_name=script_name)
            if op:
              evaluated_expr = self.evaluate_expression(op, value_map)
              value_map[variable] = evaluated_expr
              values.add(evaluated_expr)
        elif value_type == "format":
          settings, defined = get_from_dict("settings", config, self.yaml_file, silent=True, script_name=script_name)
          if defined:
            source, source_defined = get_from_dict("source", settings, self.yaml_file, silent=True, script_name=script_name)
            formatted_values = {}
            for k, v in value_map.items():
              var_cfg = self.variables.get(k, {})
              format_str = None
              if var_cfg.get("type") == "format":
                settings = var_cfg.get("settings", {})
                format_str = settings.get("format", None)
              else:
                format_str = var_cfg.get("format", None)
              formatted_values[k] = self.format_value(v, format_str)
            for name, value in formatted_values.items():
              if source_defined:
                source = str(source).replace(f"${name}", str(value)).replace(f"${{{name}}}", str(value))
            value_map[variable] = source
            values.add(source)
        elif value_type == "conversion":
          settings, defined = get_from_dict("settings", config, self.yaml_file, silent=True, script_name=script_name)
          if defined:
            from_type, _ = get_from_dict("from", settings, self.yaml_file, silent=True, script_name=script_name)
            to_type, _ = get_from_dict("to", settings, self.yaml_file, silent=True, script_name=script_name)
            source, sources_defined = get_from_dict("source", settings, self.yaml_file, silent=True, script_name=script_name)
            if sources_defined:
              source = str(source).replace("$", "").replace("{", "").replace("}", "")
            if source in value_map:
              converted = self.apply_conversion(value_map[source], from_type, to_type)
              value_map[variable] = converted
              values.add(converted)
            else:
              printc.warning(f'Source "{source}" not found for conversion variable "{variable}"', script_name)
        all_vars_values[variable] = sorted(values)

      formatted_values = {}
      for k, v in value_map.items():
        var_cfg = self.variables.get(k, {})
        format_str = None
        if var_cfg.get("type") == "format":
          settings = var_cfg.get("settings", {})
          format_str = settings.get("format", None)
        else:
          format_str = var_cfg.get("format", None)
        formatted_values[k] = self.format_value(v, format_str)
      final_template = self.template
      final_name = self.name_template
      for name, value in formatted_values.items():
        final_template = final_template.replace(f"${name}", str(value)).replace(f"${{{name}}}", str(value))
        final_name = final_name.replace(f"${name}", str(value)).replace(f"${{{name}}}", str(value))

      final_configs[final_name] = final_template

    if self.debug:
      printc.note(f"generated {len(final_configs)} configurations.", script_name)

    all_dim_vard_values = {k: sorted(set(v)) if isinstance(v, list) else [v] for k, v in dimension_vars.items()}
    all_vars_values.update(all_dim_vard_values)
    return final_configs, all_vars_values

  def generate_values_for_dim(self, var_name, var_config):
    """
    Generate the list of values for a dimension variable (bool/range/list/multiples/power_of_two).
    If it's 'function' or 'union', we skip for dimension creation here.
    """
    value_type, value_type_defined = get_from_dict("type", var_config, self.yaml_file, behavior=Key.MANTADORY, script_name=script_name)
    if value_type in ("function", "union"):
      return []

    return self.generate_values(var_config, var_name)

  def generate_values(self, config, name):
    """
    Generate values for a given variable configuration.
    
    Args:
        config (dict): Configuration settings for value generation.

    Returns:
        list: A list of generated values.
    """
    values = []

    value_type, value_type_defined = get_from_dict("type", config, self.yaml_file, parent=name, behavior=Key.MANTADORY, script_name=script_name)
    
    no_settings_var = value_type == "bool"
    settings, settings_defined = get_from_dict("settings", config, self.yaml_file, parent=name, behavior=Key.MANTADORY, silent=no_settings_var, script_name=script_name)
    if not value_type_defined or (not settings_defined and value_type != "bool"):
      return values

    if no_settings_var:
      whitelist = None
      blacklist = None
    else:
      whitelist, _ = get_from_dict("whitelist", settings, self.yaml_file, parent=name + "[settings]", silent=True, default_value=None, script_name=script_name)
      blacklist, _ = get_from_dict("blacklist", settings, self.yaml_file, parent=name + "[settings]", silent=True, default_value=None, script_name=script_name)

    if value_type == "bool":
      values = [0, 1]

    elif value_type == "range":
      from_value, from_defined = get_from_dict("from", settings, self.yaml_file, parent=name + "[settings]", behavior=Key.MANTADORY, script_name=script_name)
      to_value, to_defined = get_from_dict("to", settings, self.yaml_file, parent=name + "[settings]", behavior=Key.MANTADORY, script_name=script_name)
      step_value, _ = get_from_dict("step", settings, self.yaml_file, parent=name + "[settings]", default_value=1, silent=True, script_name=script_name)
      if to_defined and from_defined:
        values = list(range(from_value, to_value + 1, step_value))
      else:
        printc.note('You can define it like this:', script_name)
        printc.magenta("type: range:")
        printc.magenta("settings:")
        printc.magenta("  from: XXX")
        printc.magenta("  to: XXX")
        printc.magenta("  step: XXX")
        return []

    elif value_type == "power_of_two":
      from_value, from_defined = get_from_dict("from_2^", settings, self.yaml_file, parent=name + "[settings]", type=int, silent=True, script_name=script_name)
      to_value, to_defined = get_from_dict("to_2^", settings, self.yaml_file, parent=name + "[settings]", type=int, silent=True, script_name=script_name)
      if to_defined and from_defined:
        values = [2**i for i in range(int(from_value), int(to_value) + 1)]
      else:
        from_value, from_defined = get_from_dict("from", settings, self.yaml_file, parent=name + "[settings]", type=int, silent=True, script_name=script_name)
        to_value, to_defined = get_from_dict("to", settings, self.yaml_file, parent=name + "[settings]", type=int, silent=True, script_name=script_name)
        if to_defined and from_defined:
          values = [2**i for i in range(int(math.log2(from_value)), int(math.log2(to_value)) + 1)]
        else:
          printc.error('Cannot find a valid power_of_two definition for "' + name + "[settings]" + '", in "' + self.yaml_file + '".', script_name)
          printc.note('You can define it like this:', script_name)
          printc.magenta("type: power_of_two:")
          printc.magenta("settings:")
          printc.magenta("  from_2^: XXX")
          printc.magenta("  to_2^: XXX")
          printc.cyan("or ")
          printc.magenta("type: power_of_two:")
          printc.magenta("settings:")
          printc.magenta("  from: XXX")
          printc.magenta("  to: XXX")
          return []

    elif value_type == "list":
      list_values, list_defined = get_from_dict("list", settings, self.yaml_file, parent=name + "[settings]", type=list, behavior=Key.MANTADORY, script_name=script_name)
      if list_defined:
        values = list_values
      else:
        printc.note('You can define it like this:', script_name)
        printc.magenta("type: list:")
        printc.magenta("settings:")
        printc.magenta("  list: [XXX, XXX, XXX]")
        return []
    elif value_type == "multiples":
      from_value, from_defined = get_from_dict("from", settings, self.yaml_file, parent=name + "[settings]", behavior=Key.MANTADORY, script_name=script_name)
      to_value, to_defined = get_from_dict("to", settings, self.yaml_file, parent=name + "[settings]", behavior=Key.MANTADORY, script_name=script_name)
      base_value, base_defined = get_from_dict("base", settings, self.yaml_file, parent=name + "[settings]", behavior=Key.MANTADORY, script_name=script_name)
      if to_defined and from_defined and base_defined:
        values = [x for x in range(from_value, to_value + 1) if x % base_value == 0]
      else:
        printc.note('You can define it like this:', script_name)
        printc.magenta("type: multiples:")
        printc.magenta("settings:")
        printc.magenta("  from: XXX")
        printc.magenta("  to: XXX")
        printc.magenta("  base: XXX")
        return []

    # Apply whitelist/blacklist filtering
    if whitelist is not None:
      values = [v for v in values if v in whitelist]

    if blacklist is not None:
      values = [v for v in values if v not in blacklist]

    return values

  def apply_conversion(self, value, from_type, to_type):
    """
    Apply number base conversions.

    Args:
        value (str): The value to convert.
        from_type (str): What to convert from.
        to_type (str): What to convert to.

    Returns:
        str: Converted value.
    """
    try:
      if value is None:
        printc.warning(f'Cannot convert a None value from "{from_type}" to "{to_type}"', script_name)
        return value
      if from_type == "bin":
        dec_value = int(value, 2)
        if to_type == "dec":
          return str(dec_value)
        elif to_type == "hex":
          return hex(dec_value)[2:]
      elif from_type == "dec":
        dec_value = int(value)
        if to_type == "bin":
          return bin(dec_value)[2:]
        elif to_type == "hex":
          return hex(dec_value)[2:]
      elif from_type == "hex":
        dec_value = int(value, 16)
        if to_type == "bin":
          return bin(dec_value)[2:]
        elif to_type == "dec":
          return str(dec_value)
      printc.warning(f'Conversion from "{from_type}" to "{to_type}" is not supported', script_name)
    except ValueError:
      printc.error(f'Invalid value "{value}" for conversion from "{from_type}" to "{to_type}"', script_name)
    return value

  def format_value(self, value, format_string):
    """
    Format a value using the specified format string.

    Args:
        value (any): The value to format.
        format_string (str): The format string.

    Returns:
        str: Formatted value as a string.
    """
    if format_string is None:
      return str(value)
    if isinstance(value, list):
      value = "".join(str(v) for v in value)

    try:
      formatted_value = format_string % float(value)
      return formatted_value
    except (TypeError, ValueError):
      if format_string is not None:
        printc.warning(f'Cannot format value "{value}" with format "{format_string}".', script_name)
      return str(value)
